c
c     ##########################################################
c     ##                                                      ##
c     ##  subroutine nblist  --  create neighbor list         ##
c     ##                                                      ##
c     ##########################################################
c
c
c     "nblist" creates a neighbor table of switching-sites via
c     the Method of Lights due to Sullivan, Mountain and O'Connell
c     ( J. Comp Phys. Vol. 61, pp 138-153, 1985)
c
c     Author: Shawn Huston
c
c     variables:
c
c           i,j                          index for atoms
c           {ib,jb,kb,ie,je,ke,ic}       index for switching-sites
c           ik                           compressed index for switching sites
c           ksw                          number of switching sites
c
c     usage of routines COMPRS and GATHER:
c           COMPRS(v,b,u,n,ju) loads u with the XYZ locations from
c                              array v of sites indicated by array b
c
c           GATHER(v,i,u,n)    loads u with the SX-SY-SZ locations from
c                              array v of sites indicated by array i
c
c
      subroutine nblist
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'boxes.i'
      include 'cutoff.i'
      include 'neigh.i'
      include 'switch.i'
      integer i,ii,j,k,ik,kk
      integer na,nnx,nnxy,nnxyz
      integer ic,icprev,prev,ksw,kc
      integer ib,jb,kb,ie,je,ke
      integer ibu,jbu,kbu,ieu,jeu,keu
      integer nfill
      integer ibs(maxatm),jbs(maxatm),kbs(maxatm)
      integer ies(maxatm),jes(maxatm),kes(maxatm)
      integer locx(maxatm),locy(maxatm),locz(maxatm)
      integer dnbrx(maxatm),dnbrxy(maxatm),dnbrxyz(maxatm)
      integer dlnbr(maxatm)
      integer rgx(maxatm),rgy(maxatm),rgz(maxatm)
      integer bt(maxatm)
      real*8 sx(maxatm),sy(maxatm),sz(maxatm)
      real*8 rx,ry,rz,add,dist
      real*8 elapsed
c
c     temporary timing stuff
c
      call setime
c
c     get rectangle size based on cutoff and bwidth
c
      rx = chgcut + bwidth
      ry = chgcut + bwidth
      rz = chgcut + bwidth
      if ( (rx .gt. xbox2) .or. (ry .gt. ybox2)
     :                    .or. (rz .gt. zbox2) ) then
         write(6,*) 'STOP : NBLIST - box size conflict'
         stop
      end if
c
c     create compressed x-coordinate array for switching sites
c
      ik = 0
      icprev = 0
      do i = 1, n
         ic = swsite(i)
         if (ic .ne. icprev) then
            ik = ik + 1
            sx(ik) = x(ic)
            icprev = ic
            locx(ik) = ic
         end if
      end do
c
c     sort the record {sx,locx} on the key SX
c
      call sisort(sx,locx,ik)
c
c     create compressed y-coordinate array for switching sites
c
      ik = 0
      icprev = 0
      do i = 1, n
         ic = swsite(i)
         if (ic .ne. icprev) then
            ik = ik + 1
            sy(ik) = y(ic)
            icprev = ic
            locy(ik) = ic
         end if
      end do
c
c     sort the record {sy,locy} on the key SY
c
      call sisort(sy,locy,ik)
c
c     create compressed  z-coordinate array for switching sites
c
      ik = 0
      icprev = 0
      do i = 1, n
         ic = swsite(i)
         if (ic .ne. icprev) then
            ik = ik + 1
            sz(ik) = z(ic)
            icprev = ic
            locz(ik) = ic
         end if
      end do
c
c     sort the record {sz,locz} on the key SY
c
      call sisort(sz,locz,ik)
c
c     create the arrays RGY and RGZ which point from
c     array Y into SY, and Z into SZ
c
      ksw = ik
      do i = 1, ksw
         rgx(locx(i)) = i
         rgy(locy(i)) = i
         rgz(locz(i)) = i
      end do
c
c     initialize IB pointers into SX array
c
      ik = 1
      add = -xbox
      ib = 1
  1   continue
      ib = ib + 1
      if (ib .eq. ik) goto 1
  10  continue
      dist = sx(ib) + add - sx(ik)
      if ( dist .gt. -rx ) then
         ibs(locx(ik)) = ib
         goto 15
      end if
  2   continue
      ib = ib + 1
      if (ib .gt. ksw) then
      write (6,*) '2 OOPS!'
          stop
       end if
      goto 10
  15  continue
c
c     continue with remaining switching-sites
c
      do ik = 2, ksw
  20     continue
         dist = sx(ib) + add - sx(ik)
         if ( dist .gt. -rx) then
            ibs(locx(ik)) = ib
            goto 25
         end if
  4      continue
         ib = ib + 1
         if (ib .gt. ksw) then
            ib = 1
            add = 0
         end if
c
c     continue looping over ib for the current ik
c
         goto 20
  25     continue
      end do
c
c     initialize IE pointers into SX array
c
      ik = ksw
      add = xbox
      ie = ksw
  5   continue
      ie = ie - 1
      if (ie .eq. ik) goto 5
  30  continue
      if ( (sx(ie) + add - sx(ik)) .lt. rx ) then
         ies(locx(ik)) = ie
         goto 35
      end if
  6   continue
      ie = ie - 1
      if (ie .lt. 1) then
      write (6,*) '6 OOPS!'
      stop
      end if
      goto 30
  35  continue
c
c     continue with remaining switching-sites
c
      do ik = ksw-1, 1, -1
  40     continue
         dist = sx(ie) + add - sx(ik)
         if ( (sx(ie) + add - sx(ik)) .lt. rx) then
            ies(locx(ik)) = ie
            goto 45
         end if
  8      continue
         ie = ie - 1
         if (ie .lt. 1) then
            ie = ksw
            add = 0
         end if
c
c        continue looping over ie for the current ik
c
         goto 40
  45     continue
      end do
c
c     initialize JB pointers into SY array
c
      ik = 1
      add = -ybox
      jb = 1
  11  continue
      jb = jb + 1
      if (jb .eq. ik) goto 11
  50  continue
      if ( (sy(jb) + add - sy(ik)) .gt. -ry ) then
         jbs(locy(ik)) = jb
         goto 55
      end if
  12  continue
      jb = jb + 1
      if (jb .gt. ksw) then
      write (6,*) '12 OOPS!'
       stop
       end if
      goto 50
  55  continue
c
c     continue with remaining switching-sites
c
      do ik = 2, ksw
  60     continue
         if ( (sy(jb) + add - sy(ik)) .gt. -ry) then
            jbs(locy(ik)) = jb
            goto 65
         end if
  14     continue
         jb = jb + 1
         if (jb .gt. ksw) then
            jb = 1
            add = 0
         end if
c
c        continue looping over jb for the current ik
c
         goto 60
  65     continue
      end do
c
c     initialize JE pointers into SY array
c
      ik = ksw
      add = ybox
      je = ksw
  115  continue
      je = je - 1
      if (je .eq. ik) goto 115
  70  continue
      if ( (sy(je) + add - sy(ik)) .lt. ry ) then
         jes(locy(ik)) = je
         goto 75
      end if
  16   continue
      je = je - 1
      if (je .lt. 1) then
      write (6,*) '16 OOPS!'
       stop
       end if
      goto 70
  75  continue
c
c     continue with remaining switching-sites
c
      do ik = ksw-1, 1, -1
  80     continue
         if ( (sy(je) + add - sy(ik)) .lt. ry) then
            jes(locy(ik)) = je
            goto 85
         end if
  18     continue
         je = je - 1
         if (je .lt. 1) then
            je = ksw
            add = 0
         end if
c
c        continue looping over je for the current ik
c
         goto 80
  85     continue
      end do
c
c     initialize KB pointers into SZ array
c
      ik = 1
      add = -zbox
      kb = 1
  21  continue
      kb = kb + 1
      if (kb .eq. ik) goto 21
  90  continue
      if ( (sz(kb) + add - sz(ik)) .gt. -rz ) then
         kbs(locz(ik)) = kb
         goto 95
      end if
  22  continue
      kb = kb + 1
      if (kb .gt. ksw) then
         write (6,*) '22 OOPS!'
        stop
        end if
      goto 90
  95  continue
c
c     continue with remaining switching-sites
c
      do ik = 2, ksw
  100    continue
         if ( (sz(kb) + add - sz(ik)) .gt. -rz) then
            kbs(locz(ik)) = kb
            goto 105
         end if
  24     continue
         kb = kb + 1
         if (kb .gt. ksw) then
            kb= 1
            add = 0
         end if
c
c        continue looping over kb for the current ik
c
         goto 100
  105    continue
      end do
c
c     initialize KE pointers into SZ array
c
      ik = ksw
      add = zbox
      ke = ksw
  9   continue
      ke = ke - 1
      if (ke .eq. ik) goto 9
  170 continue
      if ( (sz(ke) + add - sz(ik)) .lt. rz ) then
         kes(locz(ik)) = ke
         goto 175
      end if
  116 continue
      ke = ke - 1
      if (ke .lt. 1) then
       write (6,*) '116 OOPS!'
        stop
        end if
      goto 170
  175 continue
c
c     continue with remaining switching-sites
c
      do ik = ksw-1, 1, -1
  180    continue
         if ( (sz(ke) + add - sz(ik)) .lt. rz) then
            kes(locz(ik)) = ke
            goto 185
         end if
         ke = ke - 1
         if (ke .lt. 1) then
            ke = ksw
            add = 0
         end if
c
c        continue looping over ke for the current ik
c
         goto 180
  185    continue
      end do
c
c     get intersection of X,Y, and Z atom-atom
c     neighbors based on switching-site neighbors
c     N.B. an atom will be a neighbor of itself here
c
      nfill = 0
      icprev = 0
      ik = 0
      do i = 1, n-1
         do j = 1, ksw
            bt(j) = 0
         end do
         ic = swsite(i)
c
c     get compressed switching site index for particle i
c
         if (ic .ne. icprev) ik = rgx(ic)
            ibu = ibs(locx(ik))
            ieu = ies(locx(ik))
            if (ieu .gt. ibu) then
               do ii = ibu, ieu
                  bt(ii) = 1
               end do
            else
               do ii = 1, ieu
                  bt(ii) = 1
               end do
               do ii = ibu, ksw
                  bt(ii) = 1
               end do
            end if
c
c     consolidate x-neighbors of switching-site ic
c
            call comprs(locx,bt,dnbrx,ksw,nnx)
c
c     get the SY-index of the x-neighbors
c
            call gather(rgy,dnbrx,dlnbr,ksw)
c
c     select the x-neighbor sites that are also y-neighbors
c
            jbu = jbs(locx(ik))
            jeu = jes(locx(ik))
            if (jeu .gt. jbu) then
               do j = 1, nnx
                  bt(j) = ( (jbu.le.dlnbr(j)) .and. (jeu.ge.dlnbr(j)) )
               end do
            else
               do j = 1, nnx
                  bt(j) = ( (jbu.le.dlnbr(j)) .or. (jeu.ge.dlnbr(j)) )
               end do
            end if
            bt(nnx+1) = 0
c
c     consolidate xy-neighbors of switching-site ic
c
            call comprs(dnbrx,bt,dnbrxy,nnx,nnxy)
c
c     get the SZ index of the xy-neighbors
c
            call gather(rgz,dnbrxy,dlnbr,ksw)
c
c     select the xy-neighbor sites that are also z-neighbors
c
            kbu = kbs(locx(ik))
            keu = kes(locx(ik))
            if (keu .gt. kbu) then
               do k = 1, nnxy
                  bt(k) = ( (kbu.le.dlnbr(k)) .and. (keu.ge.dlnbr(k)) )
               end do
            else
               do k = 1, nnxy
                  bt(k) = ( (kbu.le.dlnbr(k)) .or. (keu.ge.dlnbr(k)) )
               end do
            end if
c
c     consolidate xyz-neighbors of switching-site ic
c
            call comprs(dnbrxy,bt,dnbrxyz,nnxy,nnxyz)
c
c     add the neighbor atoms to particle i to the output
c     list (assumes each site has at least 1 neighbor)
c
            do j = 1, maxatm
               bt(j) = 0
            end do
            do j = 1, nnxyz
               bt(dnbrxyz(j)) = 1
            end do
            npoint(1,i) = nfill +1
c
c     fill upper triangle neighbor-list omitting diagonal
c
            do k = i+1, n
               kc = swsite(k)
               if (bt(kc) .eq. 1) then
                  nfill = nfill + 1
                  nlist(nfill) = k
               end if
            end do
            npoint(2,i) = nfill
            icprev = ic
         end do
      call getime(elapsed)
      write(6,*) 'NBLIST: Elapsed CPU time:',elapsed
      return
      end
c
c     ##########################################################
c     ##                                                      ##
c     ##  subroutine gather  --    create an array u from v   ##
c     ##                                                      ##
c     ##########################################################
c
c     'gather' loads elements from array v into array u
c     according to the index array i
c
      subroutine gather(v,i,u,n)
      implicit none
      integer j,n,i(n),u(n),v(n)
c
      do j = 1, n
         u(j) = v(i(j))
      end do
      return
      end

c
c     ##########################################################
c     ##                                                      ##
c     ##  subroutine comprs --   create an array u from v   ##
c     ##                                                      ##
c     ##########################################################
c
c     'comprs' loads elements from array v into array u
c     according to the bit array b
c
c     variable ju : number of elements in u (returned)
c
      subroutine comprs(v,b,u,n,ju)
      implicit none
      integer j,n,b(n),u(n),v(n)
      integer ib,jb,ju
c
      ju = 0
      do j = 1, n
         ib = b(j)
         do jb = 1, ib
            ju = ju + 1
            u(ju) = v(j)
         end do
      end do
      return
      end
