      SUBROUTINE CONGRA(n,par,func,tol,emin,writeout)

C     W. T. M. Mooij and B. P. van Eijck, February 1998.
C     Performs conjugate gradient or variable metric energy minimization of 
c     function FUNC with respect to N parameters PAR. 
C     Calls subroutine FUNC(n,par,g,en,isubr):
C        EN delivers the value of the function
C        G delivers the negative derivatives to the N parameters PAR
C        ISUBR=1: make pair list(s)
C             =2: use previously made pair list(s)
C     CONGRA calls FUNC first with ISUBR=1. During one minimization
C     no new pair lists must be made to ensure stability (ISUBR=2).
C     If the pair list is updated, discontinuities arise and the
C     derivatives become unreliable.

C     Input:
C        N: number of parameters
C        PAR(1..N): the parameters
C        TOL: stop earlier if RMSDER gradient < TOL

C     Output:
C        EMIN: minized energy
C        PAR(1..N): optimized parameters
      
C     The subroutine relies on a fairly independent piece of code:
C     LINE MINIMIZATION. This will search for the minimium of FUNC in a
C     given search direction ZR, using the negative derivatives G.
C     It starts with given energy EOLD and produces the minimized energy EMIN;
C     PAR is replaced by PAR1 and G by G1.
C     It starts with a given STEP in the direction ZR. To optimize STEP
C     this parameter is divided by FDIV after each minimization, but if the 
C     energy minimum is not found within STEP, this parameter is multiplied
C     by FMULT and that multiplication may be repeated NITI times.
C     For optimum performance choose the initial STEP0 in such a way that
C     NITI equals 1 or 2 in the first cycle.
C     Shanno's convergence criteria for line minimization:
C        Component of G1 in de search direction must have decreased 
C        sufficiently (criterion EPS1); the energy must have decreased
C        sufficiently (criterion EPS2) and may not be higher than the 
C        starting values at the beginning and the end of STEP.
      
c
c For general conjugate gradients or variable metric formulas see for
c example:
c [1] W.H. Press, B.P. Flannery,S.A. Teukolsky, W.T. Vetterling,
c Numerical Recipes,The art of scientific computing
c Cambridge University Press 1989, chapter 10
c 
c Parts of the algorithm (Shanno's or Shanno-Powell algorithm, Shanno's line 
c minimizations convergence critetia) are based on:
c [2] A stable rapidly converging conjugate gradient method for energy 
c minimization, S. J. Watowich, E.S. Meyer , R. Hagstrom and R. Josephs,
c J. Comp. Chemistry, 9, 650-661  
c     
c Powell-Beale and MQN formula's are from TINKER's LMQN routine
c
      IMPLICIT none
      include 'sizes.i'
      include 'keys.i'
      include 'spgr.i'
      include 'inform.i'
      include 'iounit.i'

      external func,writeout
      real*8 par(*),g(maxvar),par1(maxvar),zr(maxvar),g1(maxvar),
     +     par0(maxvar),p(maxvar),y(maxvar),deltaE(100)

      integer iperio,n,nsd,iwrite,next
      character*20 keyword
      character*80 record

      real*8 tol,emin,large,hguess

c BFGS method requires stores of Inverse Hessian, only for small systems:
      integer maxbfgs
      parameter (maxbfgs=500)
      real*8 hessin(maxbfgs,maxbfgs)
      real*8 dg(maxbfgs),hdg(maxbfgs)


      character*5 method
      
      real*8 av(maxvar),bv(maxvar),pt(maxvar),yt(maxvar),s(maxvar)
      logical reset,fine
      
      real*8 step0,shimax,fmult,fdiv,eps2,eps1,deprls,d0,
     +     eold,dold,d1g1,gnorm,step,rmsder,ya,yb,a,ea,b,eb,z,w,x
      real*8 d1gs,dnew,g1g,g1g1,pr,yy,py,pg,yg,hs,sy,sg,fac,fae,fad,
     +     sumdelta1,sumdelta2,zrnorm,gamma,gamma1,gamma2,ptyt,ytg,
     +     d1,d4,d2,d21,d3,d31,d5,d6,d7,d8,derest,emin2
      real*8 toltors,tau,dihedral,diff,tolt2

      integer nrstrt,ncyc,npow,iprint,i,j,npl,nstijg,niti,k,
     +     itors

      
      if (n.gt.maxvar) then
         write(iout,*) ' ***** Increase MAxvAR in CONGRA'
         stop
      endif
      

c Check for parameters:
c iwrite: write structure file every iwrite iterations
c iprint: print information on minimization every iprint interations
c itors: check torsions every itors iterations  
c toltors: tolerance for change in torsion
c tolfine : use large basis set from rmsgrad< tolfine on
c method: defines minimization method
c nsd: maximum number of iterations
C deprls: When the gradient-norm is satisfied, test if energy does not change
C         more than DEPRLS upon making a new pair list
C derest: When the energy changes more than DEREST upon a making a new
C         pair list, the algoritm is restarted with the gradient vector.
c large: restart with SD if rms gradient is larger than this value
C STEP0: initial step-size in  line-minimization (bv. 1.0E-5)
C        Choose such tat NITI (number of iterations in the 
c        line-minimization in the first cycle becomes 1 or 2
c shimax: make new pairlist if parameter has changed more than shimax  

      iwrite = -1
      iprint = -1
      itors = -1
      toltors = 0.0d0
      method = '     '
      nsd = -1
      derest=0.0d0      
      deprls=0.0d0
      large=0.0d0
      step0=0.0d0
      shimax=0.0d0
      hguess=0.0d0

      do i = 1, nkey
         next = 1
         record = keyline(i)
         call gettext (record,keyword,next)
         call upcase (keyword)
         if (keyword(1:9) .eq. 'PRINTOUT ') then
            read (record(next:80),*,err=20)  iprint
         else if (keyword(1:9) .eq. 'WRITEOUT ') then
            read (record(next:80),*,err=20)  iwrite
         else if (keyword(1:11) .eq. 'TORS-CHECK ') then
            read (record(next:80),*,err=20)  itors
         else if (keyword(1:9) .eq. 'TOL-TORS ') then
            read (record(next:80),*,err=20)  toltors
         else if (keyword(1:8) .eq. 'MAXITER ') then
            read (record(next:80),*,err=20)  nsd
         else if (keyword(1:5) .eq. 'BFGS ') then
            method = 'BFGS '
         else if (keyword(1:16) .eq. 'FLETCHER-REEVES ') then
            method = 'FR   '
         else if (keyword(1:14) .eq. 'POLAK-RIBIERE ') then
            method = 'PR   '
         else if (keyword(1:17) .eq. 'HESTENES-STIEFEL ') then
            method = 'HS   '
         else if (keyword(1:13) .eq. 'POWELL-BEALE ') then
            method = 'PO-BE'
         else if (keyword(1:7) .eq. 'SHANNO ') then
            method = 'SHAN '
         else if (keyword(1:14) .eq. 'SHANNO-POWELL ') then
            method = 'SH-PO'
         else if (keyword(1:5) .eq. 'LMQN ') then
            method = 'MQN  '
         else if (keyword(1:13) .eq. 'PAIRLIST-TOL ') then
            read (record(next:80),*,err=20)  deprls
         else if (keyword(1:12) .eq. 'RESTART-TOL ') then
            read (record(next:80),*,err=20)  derest
         else if (keyword(1:15) .eq. 'PAIRLIST-SHIFT ') then
            read (record(next:80),*,err=20)  shimax
         else if (keyword(1:13) .eq. 'INITIAL-STEP ') then
            read (record(next:80),*,err=20)  step0
         else if (keyword(1:15) .eq. 'LARGE-GRADIENT ') then
            read (record(next:80),*,err=20)  large
         else if (keyword(1:7) .eq. 'HGUESS ') then
            read (record(next:80),*,err=20)  hguess
         end if
 20      continue
      end do
      
c     If not read, set the defaults values:  
      if (iwrite.lt.0) iwrite = 10
      if (iprint.lt.0) iprint = 5
      if (itors.lt.0) itors = 1
      if (toltors.le.0.0d0) toltors=5.0d0
      if (method.eq.'     ') method = 'BFGS '
      if (nsd.lt.0) nsd =1000
      if (derest.le.0.0d0) derest=0.2d0      
      if (deprls.le.0.0d0) deprls=0.05d0
      if (large.le.0.0d0) large=50
      if (step0.le.0.0d0) step0=0.05
      if (shimax.le.0.0d0) shimax=0.2
      if (hguess.le.0.0d0) hguess=1.0d0

      if (nsd.eq.0) return

      if (method.eq.'BFGS ') then 
         if (n.gt.maxbfgs) then
            write(iout,25) 'Number of atoms to > MAXBFGS              '
            call fatal
         else 
            write(iout,25) 'BFGS Variable Metric                      '
         endif
      endif

      if (method.eq.'PR   ') 
     +     write(iout,25) 'Polak-Ribiere conjugate gradients          '
      if (method.eq.'SHAN ') 
     +     write(iout,25) 'Shanno conjugate gradients                 '
      if (method.eq.'HS   ') 
     +     write(iout,25) 'Hestenes-Stiefel conjugate gradients       '
      if (method.eq.'FR   ') 
     +     write(iout,25) 'Fletches-Reeves conjugate gradients        '
      if (method.eq.'SH-PO') 
     +     write(iout,25) 'Shanno conjugate gradients, Powell restarts'
      if (method.eq.'PO-BE') 
     +     write(iout,25) 'Powell-Beale conjugate gradients           '
      if (method.eq.'MQN  ') 
     +     write(iout,25) 'Memoryless BFGS-like conjugate gradients   '
      
 25   format(1x,a43)
      
      nrstrt=0
      do i = 1, nkey
         next = 1
         record = keyline(i)
         call gettext (record,keyword,next)
         call upcase (keyword)
         if (keyword(1:7) .eq. 'PERIOD ') then
            read (record(next:80),*,err=21)  nrstrt
         endif
 21      continue
      enddo

      
      if (nrstrt.eq.0) then
c Set defaults for restart criterium:        
c SH-Po and PO-BE never restart with SD direction; they do a 'Powell restart if
c the serach direction are not conjugate enough (Shanno's criterium g1g > 0.2
c g1g1), or in any case every nrstrt iterations (default n)
c BFGS only restarts with SD every nrstrt interations (defaults no restart)
c Others restart only when Shanno's criterium is not satified
         nrstrt=nsd
         if (method.eq.'PO-BE') nrstrt=n
         if (method.eq.'SH-PO') nrstrt=n
      endif

      


C     Some other default values
C     FDIV:  Divide step by FDIV, every cycle ( 1.05)
C     FMULT: Multiply step by FMULT if NITI is not zero.( 2.0)
c     eps1,eps2: values for Shanno's convergence criteria
c      
      fmult=1.2
      fdiv=1.05
      eps2=0.0001
      eps1=0.9

      iperio=0            
      d0=step0
      ncyc=0
      npow=0


 100  continue
C     Make pair list, reset reference parameters to find the shifts
      call func(n,par,g,eold,1)
      if (eold.gt.50000) then
         emin=eold
         write(iout,*) ' Unreasonable energy in CONGRA:',1,emin
         return
      endif
      do 105 i=1,n
         par0(i)=par(i)
 105  continue
      if (debug) write(iout,*) ' NEW CG PAIR LIST'

 110  continue
C     Restart search direction to the steepest descent direction
      npl=0
      dold=0.0
      do 120 i=1,n
         zr(i)=g(i)
         dold=dold+zr(i)**2
 120  continue
      
c Initialize inverse hessian for BFGS variable metric      
      
      if (method.eq.'BFGS ') then
         do i=1,n
            do j=1,n
               hessin(j,i)=0.0d0
            enddo
            hessin(i,i)=hguess
c            if (i.le.3) hessin(i,i)=10.0*hguess
         enddo
      endif

      d1g1=-dold
      gnorm=sqrt(dold)
      step=d0/gnorm
      if (eold.ge.50000) step=step*100
      rmsder=sqrt(dold/n)
      if (iprint.gt.0) print 40, ncyc,rmsder,step,eold
      if (rmsder.lt.tol) then
         emin=eold
         write(iout,*) ' ***** Gradient norm satisfied at restart CG'
         return
      endif
      
C     Start LINE MINIMIZATION.
C     Calculate energies (EA, EB) and derivatives (YA, YB) at points
C     A and B along the search direction, starting with A = 0 and B = STEP.
C     If necessary, adjust A and B to find an optimum search range.
C     Find a minimum X in the search range between A and B by determining
C     the minimum of the third-degree polynomial fixed by EA, EB, YA and YB.
 200  continue
      nstijg=0
      ya=0.0
      do 220 i=1,n
         ya=ya-zr(i)*g(i)
 220  continue
      if (ya.ge.0.) then
C        No energy gain in the search direction         
         if (debug) write(iout,*) ' ***** ya positive: ',ya
         goto 110
      endif
      a=0.0
      ea=eold
      b=step
      niti=0
 230  continue
      do 240 i=1,n
         par1(i)=par(i)+b*zr(i)
 240  continue
      call func(n,par1,g1,eb,2)
      yb=0.0
      do 250 i=1,n
         yb=yb-zr(i)*g1(i)
 250  continue

 251  continue
      
      if (debug) print 50, niti,a,b,ea,eb,ya,yb
      if (yb.lt.0.0 .and. eb.lt.ea) then
C        There must be a lower minimum further away than B, so shift and
C        double the search range.
         w=b-a
         a=b
         b=a+2*w
         ea=eb
         ya=yb
c         a=b
c         ea=eb
c         ya=yb
c         b=2.0*b
c         niti=niti+1
         goto 230
      endif
      if (eb.gt.ea+100000) then      
C        Very asymmetric energy function, decrease search range
         b=(b+a)/2.0
         if (b-a.lt.step0/10000) then
            if (iprint.gt.3) write(iout,*) ' STEP in LINMIN too small'
            call func(n,par,g,emin,1)
            if (emin.gt.50000) 
     +           write(iout,*) ' Unreasonable energy in CONGRA:',3,emin
            return
         endif
         goto 230
      endif
      
C     Now there must be a minimum X in the search range between A and B. 
C     Find it and calculate the energy there.
      if (niti.ne.0) then
         step=fmult*step
         d0=fmult*d0
      endif
      z=3.0*(ea-eb)/(b-a)+ya+yb
      w=z**2-ya*yb
      if (w.le.0.0) then
         write(iout,*) ' **** CONGRA: w < 0!!! This should never happen'
         write(iout,50) niti,a,b,ea,eb,ya,yb,w
         x=0.0
      else
         w=sqrt(w)
         x=b-(yb+w-z)/(yb-ya+2.0*w)*(b-a)
         if (debug) write(iout,50) niti,a,b,ea,eb,ya,yb,w,x/(b-a)
      endif
      
      do 274 i=1,n
         par1(i)=par(i)+x*zr(i)
         p(i)=x*zr(i)
 274  continue
c      if (debug) print 80,(par1(i),i=1,n)
      call func(n,par1,g1,emin,2)

      d1gs=0
      do i=1,n
         d1gs=d1gs-zr(i)*g1(i)
      enddo

c Check for convergence

      if ((emin.gt.eold+d1g1*x*eps2) 
     +     .or.(abs(d1gs).gt.eps1*abs(d1g1)) .or.
     +   (((emin.gt.ea) .or. (emin.gt.eb)).and.((x.gt.1.01*a).and.
     +     (x.lt.0.99*b)))) then
         if (debug) then
            if (abs(d1gs).gt.eps1*abs(d1g1)) print 72,d1gs,d1g1
            if (emin.gt.eold+d1g1*x*eps2) then
               print 70,emin,eold,d1g1,x
               print 50, niti,a,b,ea,eb,ya,yb,w,x/(b-a)
            endif
            if (emin.gt.ea .or. emin.gt.eb) print 71,emin,ea,eb
         endif

         nstijg=nstijg+1
         if ((x*d1gs).ge.0) then
            b=x
            eb=emin
            yb=d1gs
         else
            a=x
            ea=emin
            ya=d1gs
         endif
         if (nstijg.lt.5) then
C           Try a next interation within the chosen range
            goto 251
         else
C           Restart with new pair list and search direction
            if (npl.ne.0) then
               if (debug) print 35
               goto 100
            endif
C           Unless we had already restarted, then decrease step and try again
            step=step/10
            d0=d0/10
            if (step.lt.step0/10000) then
               write(iout,*) ' STEP in CONGRA too small'
               call func(n,par,g,emin,1)
               if (emin.gt.50000) write(iout,*) 
     +              ' Unreasonable energy in CONGRA:',5,emin
               return
            endif
            goto 200
         endif
      endif
      d0=d0/fdiv      
C*****end LINE MINIMIZATION
C     Now we have found a lower energy.
      
      ncyc=ncyc+1

c Write out the minimized structure to file every iwrite iterations
      if (iwrite .gt. 0) then
         if (mod(ncyc,iwrite) .eq. 0) then
            call writeout (par,ncyc,emin)
         end if
      end if

C     Calculate RMS gradient, update parameters, calc G1G for new restart
C     Calculate parameter shift (for MQN method)
      g1g1=0.0d0
      do  i=1,n
         g1g1=g1g1+g1(i)*g1(i)
         s(i)=(par1(i)-par(i))
         par(i)=par1(i)
      enddo
      rmsder=sqrt(g1g1/n)
      if (iprint.gt.0) then
         if (mod(ncyc,iprint).eq.0) print 60,niti,ncyc,rmsder,step,emin
      endif

C     Check whether we are done:
C     1. Too many cycles performed:
      if (ncyc.ge.nsd)  return
      
c     Check whether new GAMESS calculation has to be done:
      if (itors.gt.0) then
         if (mod(ncyc,itors).eq.0) then
            do i=1,nprt
               tau=dihedral(prt_tors(1,i),prt_tors(2,i),prt_tors(3,i),
     +              prt_tors(4,i))
               diff=abs(tau-tau_orig(i))
               if (diff.gt.180.d0) diff=360d0-diff
               if (rmsder.gt.0.1d0) then
                  tolt2=10.0d0*rmsder* toltors
               else
                  tolt2=toltors
               endif
               if (diff.gt.tolt2) then
                  fine=.true.
                  call writeout(par,ncyc,emin)
                  call prtgamess(fine)
                  print 81,prt_tors(1,i),prt_tors(2,i),prt_tors(3,i),
     +                 prt_tors(4,i),tau_orig(i),tau
 81               format(4i6,2f8.2)
                  print *,' TINKER Stopped to perform new GAMESS',
     +                 ' Calculation'
                  stop
               endif
            enddo
         endif
      endif
 
C     2. Gradient vector sufficiently small:
      if (rmsder.lt.tol) then
C        Check whether the pair list was still good:
         call func(n,par,g1,emin2,1)
         if (abs(emin2-emin).lt.deprls) then
            emin=emin2
            return
         endif
         if (iprint.gt.0) print 33,emin2-emin,deprls
         do 335 i=1,n
            par0(i)=par(i)
 335     continue
         iperio=iperio+1      
         deltaE(iperio)=emin2-emin
C        Return upon periodic oscillation of the energy
         if (iperio.ge.8) then
            sumdelta1=deltaE(iperio)-deltaE(iperio-2)+
     +           deltaE(iperio-4)-deltaE(iperio-6)
            sumdelta2=deltaE(iperio-1)-deltaE(iperio-3)+
     +           deltaE(iperio-5)-deltaE(iperio-7)
            if ((sumdelta1.lt.1e-6).and.(sumdelta2.lt.1e-6)) then
               write(iout,34) emin2-emin
               return
            endif
         endif
         
C        In other cases, continue
         if (abs(emin2-emin).gt.derest) then 
C           Restart search direction if the energy changed too much
            eold=emin2
            do k=1,n
               g(k)=g1(k)
            enddo
            goto 110
         else 
C           Continue normally with next search direction
            emin=emin2
            goto 361
         endif
      endif

C     Make new pair list if parameters change too much
      do 360 i=1,n
         if (abs(par(i)-par0(i)).gt.shimax) then
            call func(n,par,g1,emin2,1)
            if (iprint.gt.0) print 41, ncyc,rmsder,emin,emin2,
     +           i, par(i)-par0(i)
            do j=1,n
               par0(j)=par(j)
            enddo
            if (abs(emin2-emin).gt.derest) then 
C              New SD search vector if the energy changed too much
               eold=emin2
               do k=1,n
                  g(k)=g1(k)
               enddo
               goto 110
            else 
C              Continue normally with next search direction
               emin=emin2
               goto 361
            endif
         endif
 360  continue
C     No change in pair list necessary, continue normally

 361  continue
      if (method.eq.'PR   ') then
         pr=0.0d0         
         do i=1,n
            pr=pr+(g1(i)-g(i))*g1(i)
         enddo
      else if (method.eq.'SHAN ') then
         yy=0.0d0
         py=0.0d0
         pg=0.0d0
         yg=0.0d0
         do i=1,n
            y(i)=(g(i)-g1(i))
            yy=yy+y(i)*y(i)
            py=py+y(i)*p(i)
            pg=pg-p(i)*g1(i)
            yg=yg-y(i)*g1(i)
         enddo
      else if (method.eq.'SH-PO') then
         yy=0.0d0
         py=0.0d0
         pg=0.0d0
         yg=0.0d0
         do i=1,n
            y(i)=(g(i)-g1(i))
            yy=yy+y(i)*y(i)
            py=py+y(i)*p(i)
            pg=pg-p(i)*g1(i)
            yg=yg-y(i)*g1(i)
         enddo
      else if (method.eq.'HS   ') then
         hs=0.0d0         
         do i=1,n
            hs=hs+zr(i)*(-g1(i)+g(i))
         enddo
      else if (method.eq.'BFGS ') then
         do i=1,n
            dg(i)=-(g1(i)-g(i))            
            hdg(i)=0.0d0
         enddo
      else if (method.eq.'PO-BE') then
         yg=0.0d0         
         py=0.0d0         
         do i=1,n
            y(i)=(g(i)-g1(i))
            yg=yg-y(i)*g1(i)
            py=py+y(i)*zr(i)
         enddo
      else if (method.eq.'MQN  ') then
         yy=0.0d0
         sy=0.0d0
         yg=0.0d0
         sg=0.0d0
         do i=1,n
            s(i)=(par1(i)-par(i))
            y(i)=(g(i)-g1(i))
            yy=yy+y(i)*y(i)
            sy=sy+s(i)*y(i)
            yg=yg-y(i)*g1(i)
            sg=sg-s(i)*g1(i)
         enddo
      endif
      
C For all methods:
      g1g=0.0d0
      g1g1=0.0d0
      do 330 i=1,n
         g1g1=g1g1+g1(i)*g1(i)
c For restart criterium 
         g1g =g1g+g1(i)*g(i)
c Update derivatives;
         g(i)=g1(i)
 330  continue
      dnew=g1g1
      rmsder=sqrt(g1g1/n)
      gnorm=sqrt(g1g1)

c Update inverse hessian by BFGS algoritm
      if (method.eq.'BFGS ') then
         do j=1,n
            do i=1,n         
               hdg(i)=hdg(i) + hessin(i,j)*dg(j)
            enddo
         enddo
         fac=0.0d0
         fae=0.0d0
         do i=1,n
            fac=fac+dg(i)*zr(i)
            fae=fae+dg(i)*hdg(i)
         enddo
         fac=1.0d0/fac
         fad=1.0d0/fae
         do i=1,n
            dg(i)=fac*zr(i)-fad*hdg(i)
         enddo
         do j=1,n
            do i=1,n
               hessin(i,j)=hessin(i,j)+fac*zr(i)*zr(j)-fad*hdg(i)*hdg(j)
     +              + fae*dg(i)*dg(j)
            enddo
         enddo
      endif



      npl=npl+1
      npow=npow+1

      d1g1=0.0d0
      zrnorm=0.0d0
      reset=.false.
         
c Set new search direction:
      
c If RMS gradient is extremely high, perform another SD step
      if (rmsder.gt.large) then
c         write(iout,*) 'Another SD step as a result of large gradient'
         eold=emin
         goto 110
      endif



      if (method.eq.'PR   ' .or. method.eq.'FR   '
     +     .or. method.eq.'HS   ') then
         if (method.eq.'FR   ') gamma=dnew/dold
         if (method.eq.'PR   ') gamma=pr/dold
         if (method.eq.'HS   ') gamma=pr/hs
         do i=1,n
            zr(i)=g(i)+gamma*zr(i)
            zrnorm=zrnorm+zr(i)*zr(i)
            d1g1=d1g1-zr(i)*g(i) 
         enddo
         if (npl.eq.nrstrt) reset=.true.
         if (abs(g1g).ge. 0.2*g1g1) reset=.true.
      else if (method.eq.'SHAN ' ) then
         gamma1=(1+yy/py)*pg/py-yg/py
         gamma2=pg/py         
         do i=1,n
            zr(i)=g(i)-gamma1*p(i)+gamma2*y(i)
            zrnorm=zrnorm+zr(i)*zr(i)
            d1g1=d1g1-zr(i)*g(i) 
         enddo
         if (npl.eq.nrstrt) reset=.true.
         if (abs(g1g).ge. 0.2*g1g1) reset=.true.
      else if (method.eq.'BFGS ') then
         do i=1,n
            zr(i)=0.0d0
            do j=1,n
               zr(i)=zr(i)+hessin(i,j)*g(j)
            enddo
            zrnorm=zrnorm+zr(i)*zr(i)
            d1g1=d1g1-zr(i)*g(i)
         enddo
         if (npl.eq.nrstrt) reset=.true.
      else if (method.eq.'PO-BE') then
         gamma=yg/py
         if (abs(g1g).ge. 0.2*g1g1 .or. npl.eq.1 .or.
     +        npow.eq.nrstrt) then
            if (npl.ne.1) write(iout,*) 'POWELL restart'
            gamma1=0.0d0
            ptyt=0.0d0
            npow=0
            do i=1,n
               yt(i)=y(i)
               pt(i)=zr(i)
               ptyt=ptyt+yt(i)*pt(i)
            enddo
         else 
            ytg=0.0d0
            do i=1,n
               ytg=ytg-yt(i)*g(i)
            enddo
            gamma1=ytg/ptyt
         endif
         do i=1,n
            zr(i)=g(i)+gamma*zr(i)+gamma1*pt(i)
            zrnorm=zrnorm+zr(i)*zr(i)
            d1g1=d1g1-zr(i)*g(i)
         enddo

c If Powell-restart with Shanno's search direction:
      else if (method.eq.'SH-PO') then
         if (abs(g1g).ge. 0.2*g1g1 .or. npl.eq.1 .or.
     +        npow.eq.nrstrt) then
            if (npl.ne.1) write(iout,*) 'POWELL restart'
            npow=0
            d1=0.0d0
            d4=0.0d0
            gamma1=(1+yy/py)*pg/py-yg/py
            gamma2=pg/py         
            do i=1,n
               zr(i)=g(i)-gamma1*p(i)+gamma2*y(i)
               zrnorm=zrnorm+zr(i)*zr(i)
               d1g1=d1g1-zr(i)*g(i) 
               yt(i)=y(i)
               pt(i)=p(i)
               d1=d1+pt(i)*yt(i)
               d4=d4+yt(i)*yt(i)
            enddo 
         else
            d2=0.0d0
            d21=0.0d0
            d3=0.0d0
            d31=0.0d0
            d5=py 
            d6=0.0d0
            d7=0.0d0
            d8=pg
            do i=1,n
               d2=d2+pt(i)*y(i)
               d3=d3+yt(i)*y(i)
               d21=d21-pt(i)*g(i)
               d31=d31-yt(i)*g(i)
            enddo
            do i=1,n
               av(i)=(d1/d4)*y(i)-(d2/d4)*yt(i)+(2*d2/d1-d3/d4)*pt(i)
               bv(i)=-(d1/d4)*g(i)-(d21/d4)*yt(i)+(2*d21/d1-d31/d4)
     +              *pt(i)
            enddo 
            do i=1,n
               d6=d6+av(i)*y(i)
               d7=d7-av(i)*g(i)
            enddo
            do i=1,n
               zr(i)=-bv(i)+(d8/d5)*av(i)-((1+d6/d5)*d8/d5-d7/d5)*p(i)
               d1g1=d1g1-g(i)*zr(i)
               zrnorm=zrnorm+zr(i)*zr(i)
            enddo
         endif
      else if (method.eq.'MQN  ') then
         do i=1,n
            zr(i)= g(i) + (y(i)*sg+s(i)*yg)/sy
     +           - (1.0d0 + yy/sy)*(s(i)*sg/sy)
               d1g1=d1g1-g(i)*zr(i)
               zrnorm=zrnorm+zr(i)*zr(i)
         enddo
         if (npl.eq.nrstrt) reset=.true.
         if (abs(g1g).ge. 0.2*g1g1) reset=.true.
      endif
      
      eold=emin
      dold=dnew
      step=d0/sqrt(zrnorm)

      if (reset) goto 110

C     Next iteration: new line-minimization
      goto 200

 33   format(' Converged, but dE too large:',2f11.7,5x,
     +        'Restart with new pair list')
 34   format(' Update pair list needed (dE = ',f10.6
     +              ,') But stopped because of endless iteration')
 35   format(' More than 5 iterations in linemin.',
     +     'Restart with new pair list.')
 40   format(' RESTART CG: ncyc, rmsder, step, emin       ', 
     1     i5,f10.3,f14.8,f13.3)
 41   format('     NEW PL: ncyc, rmsder,emin1,emin2, shift',
     1     i5,f10.3,f14.3,f13.3,i5,f10.4)
 50   format(i10,2x,2f12.6,2f8.3,3e11.3,f6.2)
 60   format(7x,'niti, ncyc, rmsder, step, emin',
     1     i5,i7,f10.3,f14.8,f13.3)
 70   format(5x,'Energy increased (or decreased insufficiently)',
     +     3f13.3,e15.3)
 71   format(5x,'Energy larger than ea,eb',3f13.3)
 72   format(5x,'Derivatives decreased insufficiently:',3f13.3)
 80   format(12f10.4)
      end


