      SUBROUTINE nonbe(enb,cellder,derivs)

C     Routine to calculates nonbonded energie for crystals.
c     Based on B.P. van Eijck's UPACK program  
c     This version uses TINKER multipole electrostatics, and many other TINKER
c     routines 
      
c
C     One molecule with NRP atoms is augmented with NSPGR symmetry-related
C     ones. The nonbonded distances are selected using a cutoff criterion
C     based on the geometric centers of molecules.
C     The nonbonded energy ENB is calculated, as well as the forces on the 
C     atoms F and the derivatives of ENB
C     to the cell parameters ax, by, cz, bx, cx, cy.
C     It is possible to prepare a pair list, which can be used by other 
C     subroutines for quicker and more stable energy minimization.
      
c     Coulombic energy of polar crystals is handled by the direct summation
c     + correction approach.
c     B.P van Eijck and J. Kroon, J. Phys. Chem. B 101 (1997) 1096-1100
c

C     Output through parameters:
C     ENB: the nonbonded energy
C     CELLDER(1..6): the derivatives of ENB to BOX(1..6)
C     F(1..NR3): the derivatives of ENB to the coordinates
C     Array X,Y and Z: expanded with the symmetry-related molecules.

C     Output through common blocks (in pairlist.i)
C     /NPL/ gives pair list information for LIST:
C        NPLA: number of pairs in the list
C           The various arrays give information for the calculation of
C           nonbonded pair interactions:
C        IPLA1, IPLA2: coordinates are located in X(IPLA),Y(IPLA),Z(IPLA)
C        SHIFT1..3: fractional shifts for required symmetry operation
C        NT14list: list for 1,4 scaling
C        factorlist: list double counting factor (0.5)


      implicit none
      
      include 'sizes.i'

      include 'atoms.i'
      include 'molcul.i'
      include 'mpole.i'
      include 'vdw.i'
      include 'boxes.i'
      include 'energi.i'
      include 'deriv.i'
      include 'iounit.i'
      include 'cutoff.i'
      include 'chgpot.i'
      include 'vdwpot.i'
      include 'couple.i'
      include 'polar.i'
      include 'math.i'
      include 'units.i'
      include 'potentxtra.i'
      include 'inform.i'
      include 'spgr.i'
      include 'pairlist.i'
      include 'energixtra.i'
      include 'potent.i'
      include 'aniso.i'
      include 'charge.i'
      include 'polpot.i'


C     mxlist= maximum numbers of internal array in NONBE
      integer mxlist
      parameter(mxlist=2000)
      
      real*8 rcut,rcut2,centrex(maxatm),
     +     centrey(maxatm),centrez(maxatm)
      real*8 cellder(6),derivs(3,maxatm),field(3)
      
      
      integer i,j,m,skip(maxatm),ix,iz,k,kx,kz,i1,i2
      integer ni,nj,ii,jj,iii,jjj,nsj,inter(maxatm)
      integer ing(3*mxlist*maxatm/3)
      integer n1,n2,n3,njpt,jpt,inp(maxatm)
      real*8 enb,ax,bx,by,cx,cy,cz
      integer natm(maxatm/3),intra,nt13,nt14,ibox
      real*8 xij,yij,zij,xij0,yij0,zij0,rij2,rij
      real*8 xi,yi,zi,xr,yr,zr,factor
      real*8 ei,eikv,eik,ek,de(3,6)
      real*8 a(3,3),d(3,3,3,3)
      real*8 rpi(13),rpk(13)
      real*8 indi(3),indk(3)
      real*8 dm(3),dp(3),utu
      real*8 dmi(3,3),dmk(3,3)
      real*8 dpi(3,3),dpk(3,3)

      real*8 xpole,ypole,zpole,finner,f(4)
      real*8 indcell(3),indasym(3)
      real*8 xred(maxatm),yred(maxatm),zred(maxatm)
      real*8 xijv,yijv,zijv,rijv,rij2v,rdn,xrv,yrv,zrv
      integer iv,kv
      real*8 redi,rediv,redk,redkv,dlcsave
      integer nop,nop3,sum_msp,icj,jd,is1,is2,is3,nnj
      real*8 de1,de2,de3,y2,ddispdr,ddispdx,ddispdy,ddispdz

      logical ani_i,ani_k

      integer k1,k2

      real*8 e,dedr,qq,fc,dedx,dedy,dedz,v,uinnerpol,dumutual
      real*8 vdwtail,etail,eikd,eikr

c Set dielectric constant to intermolecular value (default 1.0):
      dlcsave=dielec
      dielec=dlcnb

c Reset pair-list
      npla=0

c Check number of multipoles:
      if ((n.ne.npole .and. n.ne.nion) .or. n.ne.nvdw) then
         write(iout,*) 
     +        'Following numbers must be identical for XXTALMIN :'
         write(iout,10) n
         write(iout,11) npole
         write(iout,13) nion
         write(iout,12) nvdw
         call fatal
      endif

 10   format('Number of atoms',i5)
 11   format('Number of multipoles',i5)
 13   format('or number of atomic charges',i5)
 12   format('Number of VDW-atoms',i5)



c We can only use one cut-off: Choose the bigger of the two
      rcut = max(chgcut,vdwcut)
      rcut2 = rcut * rcut
c
c  Get the cell parameters:
c
      ax=box(1)
      by=box(2)
      cz=box(3)
      bx=box(4)
      cx=box(5)
      cy=box(6)

c
c     zero out the list of atoms to be skipped
c
      do i = 1, n * (nspgr + 1)
         skip(i) = 0
      end do

c Fill up the unit-cell
      if (nspgr.gt.0) then
         do nop = 1,nspgr
            nop3=3*nop-3
            do  i= 1,n
               k=nop*n+i
c Expand coordinates:
               x(k) = msp(nop3+1)*x(i)+
     1              tsp(nop3+1)*ax+tsp(nop3+2)*bx+tsp(nop3+3)*cx
               y(k) = msp(nop3+2)*y(i)+
     1              tsp(nop3+2)*by+tsp(nop3+3)*cy
               z(k) = msp(nop3+3)*z(i)+tsp(nop3+3)*cz
            enddo
         enddo
      endif
        
      if (debug) then
         print*,'coordinates (after expansion):'
         do  i=1,n*(nspgr+1)
            print 91,i,x(i),y(i),z(i)
 91         format(i5,3f15.10)
         enddo
      endif

c    
c     rotate multipole components and get rotation derivatives
c     also for multipoles of equivalent positions
c
      do i = 1, npole * (nspgr + 1)
         call rotmat (i,a)
         call rotpole (i,a)
         call drotmat (i,d)
         do j = 1, 3
            do k = 1, 3
               call drotpole (i,a,d,j,k)
            end do
         end do
      end do
      
c
c Calculate y-axis orientation for anisotropic repulsion term
c
      do k = 1, naniso * (nspgr + 1)
         i = ianiso(k)
         zaxes(1,k) = x(zaxis(i)) - x(i)
         zaxes(2,k) = y(zaxis(i)) - y(i)
         zaxes(3,k) = z(zaxis(i)) - z(i)
         xaxes(1,k) = x(xaxis(i)) - x(i)
         xaxes(2,k) = y(xaxis(i)) - y(i)
         xaxes(3,k) = z(xaxis(i)) - z(i)
         call prod2(k,zaxes,xaxes,yaxes)
         y2=yaxes(1,k)*yaxes(1,k)+yaxes(2,k)*yaxes(2,k)+
     +        yaxes(3,k)*yaxes(3,k)
         leny(k)=sqrt(y2)
         leny3(k)=leny(k)*y2
         call prod2(k,yaxes,zaxes,ycrossz)
         call prod2(k,yaxes,xaxes,ycrossx)
         zminx(1,k)=x(zaxis(i)) - x(xaxis(i))
         zminx(2,k)=y(zaxis(i)) - y(xaxis(i))
         zminx(3,k)=z(zaxis(i)) - z(xaxis(i))
         call prod2(k,yaxes,zminx,ycrosszx)
      enddo

c
c     zero out multipole , polarization and Vanderwaals energy and derivatives
c
      em = 0.0d0
      ep = 0.0d0
      ev = 0.0d0
      edisp = 0.0d0
      erep = 0.0d0
      uinner=0.d0
      uinnerpol=0.0d0
      etail=0.0d0

      do i = 1, n
         do  j = 1,3
            derivs(j,i) =  0.0d0
            dev(j,i) = 0.0d0
            dem(j,i) = 0.0d0
            dep(j,i) = 0.0d0
         enddo
      enddo
      do i=1,6
         cellder(i)=0.0d0
      enddo
      do i=1,3
         field(i)=0.0d0
      enddo

c      
c Calculation of U_inner, for polar space groups only
c
      if (use_inner) then
         v = 4.0d0 * pi / (3.0d0 * ax * by * cz)
         finner = v * electric / (nspgr + 1)
C     This is twice the energy factor 2 pi / [3 V (nspgr+1)]
         xpole = 0.0d0
         ypole = 0.0d0
         zpole = 0.0d0
         if (use_mpole) then
            do ii = 1, n
               xpole = xpole + x(ii)*rpole(1,ii)*msom(1)
               ypole = ypole + y(ii)*rpole(1,ii)*msom(2)
               zpole = zpole + z(ii)*rpole(1,ii)*msom(3)
               xpole = xpole + rpole(2,ii)*msom(1)
               ypole = ypole + rpole(3,ii)*msom(2)
               zpole = zpole + rpole(4,ii)*msom(3)
            end do
         else if (use_charge) then
            do ii = 1, nion
               xpole = xpole + x(ii)*pchg(ii)*msom(1)
               ypole = ypole + y(ii)*pchg(ii)*msom(2)
               zpole = zpole + z(ii)*pchg(ii)*msom(3)
            enddo
         endif
            
         if (verbose) print 71,xpole*debye,ypole*debye,zpole*debye
 71      format(' Celldipole',2x,3f8.3)
         f(2)=finner*xpole
         f(3)=finner*ypole
         f(4)=finner*zpole
      
c        U_inner term itself:
         uinner = uinner - f(2) * xpole/2
         uinner = uinner - f(3) * ypole/2
         uinner = uinner - f(4) * zpole/2
c This term is the long-range correction to the electrostatic energy
         em = em + uinner

c Derivatives of Uinner to the coordinates:         
c dU_inner/dx_i, dU_inner/dy_i, dU_inner/dz_i
         if (use_mpole) then
            do i = 1 , n 
               iz = zaxis(i)
               ix = xaxis(i)
               do  j = 1,3
                  derivs(j,i) =  derivs(j,i)-f(j+1)*rpole(1,i) * msom(j)
                  do k=1,3
                     derivs(j,i) =  derivs(j,i) - f(k+1) * 
     +                    dpole(k+1,1,j,i) * msom(k)
                     derivs(j,iz) = derivs(j,iz) - f(k+1) * 
     +                    dpole(k+1,2,j,i) * msom(k)
                     derivs(j,ix) = derivs(j,ix) - f(k+1) *
     +                    dpole(k+1,3,j,i) * msom(k)
                  enddo
               enddo 
            enddo
         else if (use_charge) then
            do i = 1 , n 
               do  j = 1,3
                  derivs(j,i) =  derivs(j,i)-f(j+1)*pchg(i) * msom(j)
               enddo
            enddo 
         endif          
c Derivatives of Uinner to cell-volume           
c dUinner/dax, dUinner/day, dUinner/daz, 
c
         do i = 1,3
            cellder(i) = -uinner/box(i)
         enddo
         field(1) = xpole * v
         field(2) = ypole * v
         field(3) = zpole * v
      endif
c End Uinner term
      
c      
c Calculation of long-range tail corrections to Vanderwaals interactions
c Initialize cell-derivatives.
c
      if (use_tail) then
         etail=vdwtail()
         do i=1,3
            cellder(i)=cellder(i)-etail/box(i)
         enddo
         ev=etail
         edisp=etail
      endif
c     
c     compute the induced dipoles at each atom
c
c Field contains the field due to the charge-distribtion on the cut
c -off sphere. This is the same kind of long-range correction that causes the 
c u-inner term for the electrostatic energy.
c

      call xinducenonbe(field)
c
c Now we calculate the u-inner correction for the polarization
c
      if (use_inner) then
         do i=1,3
            indasym(i)=0.0d0
         enddo
         do ii = 1, n
            do i=1,3
               indasym(i)= indasym(i)+uind(i,ii)
            enddo
         enddo
         do i=1,3
            indcell(i)=indasym(i)*msom(i)
         enddo
         if (verbose) print 72,indcell(1)*debye,indcell(2)*ypole*debye,
     +        indcell(3)*debye
 72      format(' Induced   ',2x,3f8.3)

         
         dumutual = 0.0d0
c
c The polarization energy calculated with empik1 does not yet contain the
c interaction of the induced dipoles with the u-inner field.
c dumutual is a term for the induced dipole-induced dipole gradients
c
         do i=1,3
            uinnerpol= uinnerpol -0.5d0*electric*indasym(i)*
     +           field(i)
            dumutual = dumutual + electric*v*indcell(i)*indasym(i)
         enddo

         if (use_mpole) then
            do i = 1 , n 
               iz = zaxis(i)
               ix = xaxis(i)
               do  j = 1,3
                  dep(j,i) =  dep(j,i)- v*electric*
     +                 indasym(j)* rpole(1,i) * msom(j)
                  do k=1,3
                     dep(j,i) =  dep(j,i) - v*electric * 
     +                    indasym(k) * dpole(k+1,1,j,i) * msom(k)
                     dep(j,iz) = dep(j,iz) - v*electric * 
     +                    indasym(k) * dpole(k+1,2,j,i) * msom(k)
                     dep(j,ix) = dep(j,ix) - v*electric * 
     +                    indasym(k) * dpole(k+1,3,j,i) * msom(k)
                  enddo
               enddo 
            enddo
         endif
         do i = 1,3
            cellder(i) = cellder(i) - 2.0d0* uinnerpol/box(i)
            if (poltyp.eq.'MUTUAL') 
     +           cellder(i) = cellder(i) + 0.5d0*dumutual/box(i) 
         enddo
c This term is the long-range correction to the polarization energy
         ep = ep + uinnerpol
      endif

c
c Calculate reduced atomic coordinates for VDW interaction, also for equivalent
c positions 
c
      do i = 1, n
         do j= 0,nspgr
c            i = ivdw(k)
            iv = ired(i) + j*n
            rdn = kred(i)
            ii=i + j*n
            xred(ii) = rdn*(x(ii)-x(iv)) + x(iv)
            yred(ii) = rdn*(y(ii)-y(iv)) + y(iv)
            zred(ii) = rdn*(z(ii)-z(iv)) + z(iv)
         enddo  
c The whole routine is looping is over the atoms, which all should have VDW and
c multipole interactions. Else the algoritm does not work.
         if (ipole(i).ne.i) then 
            if (iion(i).ne.i) then
               print *,'ipole/iion(i) <> i, this should not happen'
               call fatal
            endif
         else if (ivdw(i).ne.i) then 
            print *,'ivdw(i) <> i, this should not happen'
            call fatal
         endif
      end do

c      
c If a polarization is used, then xinduce has just made a new pairlist:
      if (use_polar) goto 666
      

C*****CALCULATE AND STORE TEMPORARILY THE CENTERS OF GEOMETRY
C*****OF THE CHARGE GROUPS
c***** Charge-groups are complete molecules 
      
      do i = 1, nmol
         natm(i) =  0
         centrex(i)=0.0d0
         centrey(i)=0.0d0
         centrez(i)=0.0d0
      enddo

      do j = 1, n
         k = molcule(j)
         centrex(k) = centrex(k) + x(j)
         centrey(k) = centrey(k) + y(j)
         centrez(k) = centrez(k) + z(j)
         natm(k) = natm(k) + 1
      enddo
      
      do i= 1 , nmol
         centrex(i) = centrex(i)/natm(i)
         centrey(i) = centrey(i)/natm(i)
         centrez(i) = centrez(i)/natm(i)
      enddo
      
c     centres of geometry of equivalent positions:  
      if (nspgr.gt.0) then
         do nop = 1,nspgr
            nop3=3*nop-3
            do  i= 1,nmol
               j=nop*nmol+i
               centrex(j) = msp(nop3+1)*centrex(i)+
     1              tsp(nop3+1)*ax+tsp(nop3+2)*bx+tsp(nop3+3)*cx
               centrey(j) = msp(nop3+2)*centrey(i)+
     1              tsp(nop3+2)*by+tsp(nop3+3)*cy
               centrez(j) = msp(nop3+3)*centrez(i)+tsp(nop3+3)*cz
            enddo
         enddo
      endif
      
C     Loop over all molecules NI in asymmetric unit
      do 400 ni = 1,nmol
         if (debug) print 80,ni
 80      format(/,' Start calculation for charge group',i4)

C        Loop over molecules NJ , including equivalent positionsin
c        Create a list of translations needed to find all charge-group
c        combinations within RCUT. The translations are stored in array
C           ING in groups (N1, N2, N3) for every translation 1...NSJ, 
C           starting at NJPT; NSJ is stored in INP(NJ). 
         do 200 nj = ni , nmol * (nspgr+1)
            njpt = (nj-1)*mxlist*3
C NOTE:          Intrabonds are the first in the list
c          No translation within the molecule
            xij0 = centrex(ni) - centrex(nj)
            yij0 = centrey(ni) - centrey(nj)
            zij0 = centrez(ni) - centrez(nj)
c            print*,xij0,yij0,zij0
            if (xij0**2+yij0**2+zij0**2.lt.rcut2) then
               nsj = 1
c               print 155,nsj,0,0,0,xij0,yij0,zij0,
c     1           sqrt(xij0**2+yij0**2+zij0**2)
               inter(nj) = 0
               do  m = 1,3
                  ing(njpt+m) = 0
               enddo
            else
               nsj=0
               inter(nj)=1
            endif
            n3 = int((rcut - zij0)/cz)
            zij = zij0 + n3*cz
            n2 = int((rcut - yij0 - n3*cy)/by)
            yij = yij0 + n2*by + n3*cy
            n1 = int((rcut - xij0 - n2*bx - n3*cx)/ax)
            xij=xij0 + n1*ax + n2*bx + n3*cx

C           Look for  charge-groups within RCUT
 150        continue
            if (n1.eq.0 .and. n2.eq.0 .and. n3.eq.0) goto 160
            rij2 = xij**2 + yij**2 + zij**2
            if (rij2.lt.rcut2) then
               nsj = nsj + 1
c               print 155,nsj,n1,n2,n3,xij,yij,zij,sqrt(rij2)
 155           format(i5,3x,3i5,10f8.3)
               if (nsj.gt.mxlist) then
                  print*,' ***** mxlist te klein'
                  write(16,*) ' ***** mxlist te klein'
                  stop
               endif
               jpt = njpt + 3*(nsj-1)
               ing(jpt+1) = n1
               ing(jpt+2) = n2
               ing(jpt+3) = n3
            endif
 160        continue
            n1 = n1-1
            xij = xij-ax
 170        if (xij.lt.-rcut) then
               n2 = n2-1
               yij = yij - by
 180           if (yij.lt.-rcut) then
                  n3 = n3-1
                  zij = zij-cz
                  if (zij.lt.-rcut) goto 190
                  n2 = int((rcut - yij0 - n3*cy)/by)
                  yij = yij0 + n2*by + n3*cy
                  goto 180
               endif
               n1 = int((rcut - xij0 - n2*bx - n3*cx)/ax)
               xij = xij0 + n1*ax + n2*bx + n3*cx
               goto 170
            endif
            goto 150
 190        inp(nj) = nsj
c            print*,ni,nj,nsj
 200     continue
C        end of making list ING
        
c         
C        Loop over atoms II (1..N) within molecule NI.
c  First and last atom of molecule ni:
         do 390 iii = imol(1,ni), imol(2,ni)
            ii = kmol(iii)     
            xi = x(ii)
            yi = y(ii)
            zi = z(ii)
            iz = zaxis(ii)
            ix = xaxis(ii)

C For "reduced" coordinates
            iv = ired(ii)
            redi = kred(ii)
            rediv = 1.0d0 - redi
            
c
C For excluded neighbours the vdw settings are used: 
c if vdw are not scaled, charge is not scaled.
c if scaling, then VDW and Charge will be scaled by their own scaling values
c 1-4 interactions are never scaled for multipole interactions
c only for VDW , or CHARGE actions
c
            do j = 1, n12(ii)
               skip(i12(j,ii)) = ii * vdw12use
            end do
            do j = 1, n13(ii)
               skip(i13(j,ii)) = ii * vdw13use
            end do
            do j = 1, n14(ii)
               skip(i14(j,ii)) = ii * vdw14use
            end do

            do j = 1, 13
               rpi(j) = rpole(j,ii)
            end do
            do j = 1, 3
               indi(j) = uind(j,ii)
            end do
            
C           Loop over charge-groups NJ (including equivalent positions).
            DO 300 NJ = ni , nmol * (nspgr + 1)
               IF (INP(NJ).EQ.0) GOTO 300
               njpt = (nj-1)*mxlist*3

C              ICJ is the number of the equivalent molecule, 0..NSPGR.
C              ICJ=0 denotes a distance within the asymmetric unit, which
C              is intramolecular if this contains one molecule

               ICJ=(NJ-1)/nmol

C              Loop over atoms JJ (1..NRP) within charge-group NJ.
               
               nnj=nj - nmol*icj
               do 280 jjj = imol(1,nnj) , imol(2,nnj)

c Loop is only over J>I atoms
                  
                  if (jjj.lt.iii) goto 280
c                  
c JJ is the number in the asymmetric unit numbering, k is the numbering in the
c unit cell.
                  jj = kmol(jjj)
                  k = jj + icj * n

c Axes-defining atoms in the asymmetric unit numbering:
                  kz = zaxis(jj)
                  kx = xaxis(jj)
C For "reduced" coordinates
                  kv = ired(jj)

 92               format(7x,'charge groups NI, NJ:',2i4,
     +                 8x,'atoms II, JJ,  JJ-exp',3i4)
                  if (debug) print 92,ni,nj,ii,jj,k

                  jd=3*icj
                  if (jd.eq.0) then
                     is1=1
                     is2=1
                     is3=1
                     de1=0
                     de2=0
                     de3=0
                  else
                     is1=msp(jd-2)
                     is2=msp(jd-1)
                     is3=msp(jd)
                     de1=-tsp(jd-2)
                     de2=-tsp(jd-1)
                     de3=-tsp(jd)
                  endif 
                  
c                 Search for excluded neighbours  
                  intra=0
                  if (icj.eq.0 .and. inter(nj).eq.0) intra=1

                  nt13=0
                  nt14=0
c
c Excluded neighbours only when intramolecular (intra flag = 1)
c 
                  if (intra.eq.1) then
                     if (skip(k) .eq. ii )  nt13=1
                     if (skip(k) .eq.-ii )  nt14=1
                  endif
                  
                  do j = 1, 13
                     rpk(j) = rpole(j,k)
                  end do

                  do j = 1, 3
                     indk(j) = uind(j,k)
                  end do
                  
                  factor=1.0d0 
                  if (jj.eq.ii) factor = 0.5d0
                     
                  xr = x(ii) - x(k) 
                  yr = y(ii) - y(k)
                  zr = z(ii) - z(k)
C For "reduced" coordinates
                  xrv = xred(ii) - xred(k)
                  yrv = yred(ii) - yred(k)
                  zrv = zred(ii) - zred(k)
                  
C                    Start the next cell translation from ING
                  do 260 ibox=1,inp(nj)
                     n1=ing(njpt+(ibox-1)*3+1)
                     n2=ing(njpt+(ibox-1)*3+2)
                     n3=ing(njpt+(ibox-1)*3+3)
c If no single intra-molecular term is to be used: skip
c intra.eq.1 means we are not in an equivalent/translated position
c test for molcule(ii).eq.molcule(jj) for multiple molecules in the A.U.
                     if (intra.eq.1 .and. molcule(ii).eq.molcule(jj)
     +                    .and. nointranonbe) goto 250
c
c If atoms are the same in cell-expanded numbering and translation is zero:
c skip
                     if (n1.eq.0 .and. n2.eq.0 .and. n3.eq.0 
     +                    .and. ii.eq.k)  goto 250

c                    If it are excluded neighbours: skip
                     if (nt13.eq.1) goto 250

                     xij=xr+n1*ax+n2*bx+n3*cx
                     yij=yr+n2*by+n3*cy
                     zij=zr+n3*cz
                     rij2=xij**2+yij**2+zij**2
                     rij=sqrt(rij2)
c
c                    For reduced coordinates
c
                     if ((ii.ne.iv) .or. (jj.ne.kv)) then 
                        xijv=xrv+n1*ax+n2*bx+n3*cx
                        yijv=yrv+n2*by+n3*cy
                        zijv=zrv+n3*cz
                        rij2v=xijv**2+yijv**2+zijv**2
                        rijv=sqrt(rij2v)
                     else 
                        xijv=xij
                        yijv=yij
                        zijv=zij
                        rij2v=rij2
                        rijv=rij
                     endif


                     if (rij2.lt.1.0e-6) goto 250
C                    Create atom pair list for LIST routine
                     npla=npla+1
                     if (npla.gt.maxlist) then
                        print*, ' ***** NONBE: Increase maxlist'
                        stop
                     endif
                     ipla1(npla)=ii
                     ipla2(npla)=k
                     shift1(npla)=n1
                     shift2(npla)=n2
                     shift3(npla)=n3
                     factorlist(npla)=factor
                     nt14list(npla)=nt14
c                     
C Increment the multipole energy and derivatives
c
                     eik=0.0d0
                     ei=0.0d0
                     ek=0.0d0
                     if (use_mpole) then 
                        call empik1 (ii,k,-xij,-yij,-zij,rij,rij2,
     +                       rpi,rpk,indi,indk,eik,ei,ek,dm,dmi,dmk,
     +                       dp,dpi,dpk,utu)
                        em = em + eik * factor
                        if (factor.ne.1.0d0) then
                           do i1 = 1,3
                              dm(i1) = dm(i1) * factor
                              do i2 = 1,3
                                 dmi(i1,i2)=dmi(i1,i2) * factor
                                 dmk(i1,i2)=dmk(i1,i2) * factor
                              enddo
                           enddo
                        endif
                        dem(1,ii) = dem(1,ii) - dm(1) + dmi(1,1)
                        dem(2,ii) = dem(2,ii) - dm(2) + dmi(1,2)
                        dem(3,ii) = dem(3,ii) - dm(3) + dmi(1,3)
                        dem(1,iz) = dem(1,iz) + dmi(2,1)
                        dem(2,iz) = dem(2,iz) + dmi(2,2)
                        dem(3,iz) = dem(3,iz) + dmi(2,3)
                        dem(1,ix) = dem(1,ix) + dmi(3,1)
                        dem(2,ix) = dem(2,ix) + dmi(3,2)
                        dem(3,ix) = dem(3,ix) + dmi(3,3)
                        dem(1,jj) = dem(1,jj) +is1*(dm(1) + dmk(1,1))
                        dem(2,jj) = dem(2,jj) +is2*(dm(2) + dmk(1,2))
                        dem(3,jj) = dem(3,jj) +is3*(dm(3) + dmk(1,3))
                        dem(1,kz) = dem(1,kz) + is1 * dmk(2,1)
                        dem(2,kz) = dem(2,kz) + is2 * dmk(2,2)
                        dem(3,kz) = dem(3,kz) + is3 * dmk(2,3)
                        dem(1,kx) = dem(1,kx) + is1 * dmk(3,1)
                        dem(2,kx) = dem(2,kx) + is2 * dmk(3,2)
                        dem(3,kx) = dem(3,kx) + is3 * dmk(3,3)

                        dedx=-dm(1)+dmi(1,1) + dmi(2,1) + dmi(3,1)
                        dedy=-dm(2)+dmi(1,2) + dmi(2,2) + dmi(3,2)
                        dedz=-dm(3)+dmi(1,3) + dmi(2,3) + dmi(3,3)
                        cellder(1)=cellder(1) + (n1+de1) * dedx
                        cellder(2)=cellder(2) + (n2+de2) * dedy
                        cellder(3)=cellder(3) + (n3+de3) * dedz
                        cellder(4)=cellder(4) + (n2+de2) * dedx
                        cellder(5)=cellder(5) + (n3+de3) * dedx
                        cellder(6)=cellder(6) + (n3+de3) * dedy

 888                    format (12x,3i3,2x,4f8.3,3x,
     +                    f8.3,f8.3)
 886                    format ('RPI ',13f7.3)
 887                    format ('RPK ',13f7.3)

                     
c
c     increment the polarization energy and derivative expressions
c

                        if (use_polar) then
                           ep = ep + (ei + ek) * factor
                           if (factor.ne.1.0d0) then
                              do i1 = 1,3
                                 dp(i1) = dp(i1) * factor
                                 do i2 = 1,3
                                    dpi(i1,i2)=dpi(i1,i2) * factor
                                    dpk(i1,i2)=dpk(i1,i2) * factor
                                 enddo
                              enddo 
                           endif
                           
                           dep(1,ii) = dep(1,ii) - dp(1) + dpi(1,1)
                           dep(2,ii) = dep(2,ii) - dp(2) + dpi(1,2)
                           dep(3,ii) = dep(3,ii) - dp(3) + dpi(1,3)
                           dep(1,iz) = dep(1,iz) + dpi(2,1)
                           dep(2,iz) = dep(2,iz) + dpi(2,2)
                           dep(3,iz) = dep(3,iz) + dpi(2,3)
                           dep(1,ix) = dep(1,ix) + dpi(3,1)
                           dep(2,ix) = dep(2,ix) + dpi(3,2)
                           dep(3,ix) = dep(3,ix) + dpi(3,3)
                           dep(1,jj) = dep(1,jj) + is1*( dp(1)+dpk(1,1))
                           dep(2,jj) = dep(2,jj) + is2*( dp(2)+dpk(1,2))
                           dep(3,jj) = dep(3,jj) + is3*( dp(3)+dpk(1,3))
                           dep(1,kz) = dep(1,kz) + is1 * dpk(2,1)
                           dep(2,kz) = dep(2,kz) + is2 * dpk(2,2)
                           dep(3,kz) = dep(3,kz) + is3 * dpk(2,3)
                           dep(1,kx) = dep(1,kx) + is1 * dpk(3,1)
                           dep(2,kx) = dep(2,kx) + is2 * dpk(3,2)
                           dep(3,kx) = dep(3,kx) + is3 * dpk(3,3)
                     
                           dedx= -dp(1) + dpi(1,1) + dpi(2,1) + dpi(3,1)
                           dedy= -dp(2) + dpi(1,2) + dpi(2,2) + dpi(3,2)
                           dedz= -dp(3) + dpi(1,3) + dpi(2,3) + dpi(3,3)
                           cellder(1)=cellder(1) + (n1+de1) * dedx
                           cellder(2)=cellder(2) + (n2+de2) * dedy
                           cellder(3)=cellder(3) + (n3+de3) * dedz
                           cellder(4)=cellder(4) + (n2+de2) * dedx
                           cellder(5)=cellder(5) + (n3+de3) * dedx
                           cellder(6)=cellder(6) + (n3+de3) * dedy
                        endif
                     endif
                  
c Increment Charge-Charge energy and derivatives                  
                     if (use_charge) then
                        if (nt14.eq.1) then
                           fc=factor/chgscale
                        else
                           fc=factor
                        endif
                        
                        qq=fc*pchg(ii)*pchg(jj)*electric / dielec
                        e=qq/rij
                        dedr=-qq/rij2
c
c     form the chain rule terms for derivative expressions
c
                        dedr = dedr / rij
                        dedx = dedr * xij
                        dedy = dedr * yij
                        dedz = dedr * zij
c
c     increment the 'multipole' energy and derivative expressions
c    for the charge-charge interaction case.
                        em = em + e
                        dem(1,ii) = dem(1,ii) + dedx
                        dem(2,ii) = dem(2,ii) + dedy
                        dem(3,ii) = dem(3,ii) + dedz
                        dem(1,jj) = dem(1,jj) - is1 * dedx
                        dem(2,jj) = dem(2,jj) - is2 * dedy
                        dem(3,jj) = dem(3,jj) - is3 * dedz
                        cellder(1)=cellder(1) + (n1+de1) * dedx
                        cellder(2)=cellder(2) + (n2+de2) * dedy
                        cellder(3)=cellder(3) + (n3+de3) * dedz
                        cellder(4)=cellder(4) + (n2+de2) * dedx
                        cellder(5)=cellder(5) + (n3+de3) * dedx
                        cellder(6)=cellder(6) + (n3+de3) * dedy
                     endif
                     
c Vanderwaals energy:
                     eikv=0.0
                     call evdwpair1(ii,k,jj,xijv,yijv,zijv,rij2v,rijv,
     +                    rij2,rij,ddispdr,eikv,eikd,eikr,
     +                    de,ani_i,ani_k,nt14)

                     ev = ev + eikv* factor
                     edisp = edisp+eikd*factor
                     erep = erep+eikr*factor
                     
                     if (factor.ne.1.0d0) then
                        ddispdr=ddispdr*factor
                        do i1=1,6
                           do i2=1,3
                              de(i2,i1) = de(i2,i1)*factor
                           enddo
                        enddo
                     endif

                     if (ii.eq.iv) then
                        dev(1,ii) = dev(1,ii) + de(1,1)
                        dev(2,ii) = dev(2,ii) + de(2,1)
                        dev(3,ii) = dev(3,ii) + de(3,1)
                        dedx = de(1,1)
                        dedy = de(2,1)
                        dedz = de(3,1)
                        if (ani_i) then
                           dev(1,iz) = dev(1,iz) + de(1,2)
                           dev(2,iz) = dev(2,iz) + de(2,2)
                           dev(3,iz) = dev(3,iz) + de(3,2)
                           dev(1,ix) = dev(1,ix) + de(1,3)
                           dev(2,ix) = dev(2,ix) + de(2,3)
                           dev(3,ix) = dev(3,ix) + de(3,3)
                           dedx=dedx + de(1,2) + de (1,3)
                           dedy=dedy + de(2,2) + de (2,3)
                           dedz=dedz + de(3,2) + de (3,3)
                        endif
                     else
                        dedx = de(1,1)
                        dedy = de(2,1)
                        dedz = de(3,1)
                        dev(1,ii) = dev(1,ii) + de(1,1) * redi
                        dev(2,ii) = dev(2,ii) + de(2,1) * redi
                        dev(3,ii) = dev(3,ii) + de(3,1) * redi
                        dev(1,iv) = dev(1,iv) + de(1,1) * rediv
                        dev(2,iv) = dev(2,iv) + de(2,1) * rediv
                        dev(3,iv) = dev(3,iv) + de(3,1) * rediv
                     end if
                     if (jj .eq. kv) then
                        dev(1,jj) = dev(1,jj) + is1 * de(1,6)
                        dev(2,jj) = dev(2,jj) + is2 * de(2,6)
                        dev(3,jj) = dev(3,jj) + is3 * de(3,6)
                        if (ani_k) then
                           dev(1,kz) = dev(1,kz) + is1*de(1,4)
                           dev(2,kz) = dev(2,kz) + is2*de(2,4)
                           dev(3,kz) = dev(3,kz) + is3*de(3,4)
                           dev(1,kx) = dev(1,kx) + is1*de(1,5)
                           dev(2,kx) = dev(2,kx) + is2*de(2,5)
                           dev(3,kx) = dev(3,kx) + is3*de(3,5)
                        endif
                     else
                        redk = kred(jj)
                        redkv = 1.0d0 - redk
                        dev(1,jj) = dev(1,jj) + is1 * de(1,6)*redk
                        dev(2,jj) = dev(2,jj) + is2 * de(2,6)*redk
                        dev(3,jj) = dev(3,jj) + is3 * de(3,6)*redk
                        dev(1,kv) = dev(1,kv) + is1 * de(1,6) * redkv
                        dev(2,kv) = dev(2,kv) + is2 * de(2,6) * redkv
                        dev(3,kv) = dev(3,kv) + is3 * de(3,6) * redkv
                     end if
c
c Derivatives for dispersion calculated without reduced coordinates  
c (used in Aexp(-br)-C/r6)
c
                     if (ddispdr.ne.0.0d0) then
                        ddispdr=ddispdr/rij
                        ddispdx=ddispdr*xij
                        ddispdy=ddispdr*yij
                        ddispdz=ddispdr*zij
                        dev(1,ii) = dev(1,ii) + ddispdx
                        dev(2,ii) = dev(2,ii) + ddispdy
                        dev(3,ii) = dev(3,ii) + ddispdz
                        dev(1,jj) = dev(1,jj) - is1 * ddispdx
                        dev(2,jj) = dev(2,jj) - is2 * ddispdy
                        dev(3,jj) = dev(3,jj) - is3 * ddispdz
                        dedx = dedx + ddispdx
                        dedy = dedy + ddispdy
                        dedz = dedz + ddispdz
                     endif

                     cellder(1)=cellder(1) + (n1+de1) * dedx
                     cellder(2)=cellder(2) + (n2+de2) * dedy
                     cellder(3)=cellder(3) + (n3+de3) * dedz
                     cellder(4)=cellder(4) + (n2+de2) * dedx
                     cellder(5)=cellder(5) + (n3+de3) * dedx
                     cellder(6)=cellder(6) + (n3+de3) * dedy

 889                 format (12x,3i3,2x,4f8.3,3x,
     +                    f8.3,2e13.4,12x,8f8.3)
                     
 250                 continue
                     if (ibox.eq.1) then
                        intra=0
                        nt13=0
                        nt14=0
                     endif
C             End of loop over cell-translations
 260              continue
 280           continue
 300        continue
 390     continue
 400  continue


      enb= em + ep + ev
      if (verbose) then
         if (em .ne. 0.0) print 111,'E-multipole     ',em
         if (uinner .ne. 0.0) print 111,'incl. U2-inner  ',uinner
         if (ep .ne. 0.0) print 111,'E-polarisation  ',ep
         if (uinnerpol .ne. 0.0) print 111,'incl. U2-pol    ',uinnerpol
         if (ev .ne. 0.0) print 111,'E-Vanderwaals   ',ev
         if (etail .ne. 0.0) print 111,'incl. E-tail    ',etail
      endif
      if (verbose) print 111,'E-nonbe         ',enb

 111  format(a17,f15.10)
      do i = 1, n
         do j = 1, 3
            derivs(j,i) = derivs(j,i) + dem(j,i) + dev(j,i) + dep(j,i)
         end do
      end do
c      
c Reset dielectric constant to old value for intramolecular interaction:
c (This is strange, but it allows for 'correct' MM3 for intramolecular
c interactions,including it's dielectric constant of 1.5
c 
      dielec=dlcsave

c      print 298, (der(i),i=1,6)
c      print 299, (f(i),i=1,nr3)
c      write(93,297) ' Energie Nonbe: ',eel,elj,enb
c      write(93,298) (der(i),i=1,6)
c      write(93,299) (f(i),i=1,nr3)
 297  format(' Energie Nonbe: ',5f12.5,5x,'Mpole,Pola,Vdw,Uinner,Enb')
c  298 format(' Der: ',6f9.1)
c  299 format(9f8.1)
      return
      
c
c      
c For polarizable model: INDUCE has just created a pairlist, so we use this
c pairlist to calculate the nonbonded energy.
c
 666  continue


      do ni=1,npla
         ii=ipla1(ni)
c k  is the atom number in cell-expansion
         k=ipla2(ni)
         icj=(k-1)/n
c jj is the atom numbern asymmetric unit
         jj=k-icj*n
         jd=3*icj

         if (jd.eq.0) then
            is1=1
            is2=1
            is3=1
            de1=0
            de2=0
            de3=0
         else
            is1=msp(jd-2)
            is2=msp(jd-1)
            is3=msp(jd)
            de1=-tsp(jd-2)
            de2=-tsp(jd-1)
            de3=-tsp(jd)
         endif 
         n1=shift1(ni)
         n2=shift2(ni)
         n3=shift3(ni)
         factor=factorlist(ni)
         nt14=nt14list(ni)
         xi = x(ii)
         yi = y(ii)
         zi = z(ii)
         iz = zaxis(ii)
         ix = xaxis(ii)

C For "reduced" coordinates
         iv = ired(ii)
         redi = kred(ii)
         rediv = 1.0d0 - redi
         do j = 1, 13
            rpi(j) = rpole(j,ii)
         end do
         do j = 1, 3
            indi(j) = uind(j,ii)
         end do
         kz = zaxis(jj)
         kx = xaxis(jj)
         do j = 1, 13
            rpk(j) = rpole(j,k)
         end do

         do j = 1, 3
            indk(j) = uind(j,k)
         end do


         xr = x(ii) - x(k) 
         yr = y(ii) - y(k)
         zr = z(ii) - z(k)
         rij2=0.0
         xij=xr+n1*ax+n2*bx+n3*cx
         yij=yr+n2*by+n3*cy
         zij=zr+n3*cz
         rij2=xij**2+yij**2+zij**2
         rij=sqrt(rij2)

C For "reduced" coordinates
         kv = ired(jj)
         if ((ii.ne.iv) .or. (jj.ne.kv)) then 
            xrv = xred(ii) - xred(k)
            yrv = yred(ii) - yred(k)
            zrv = zred(ii) - zred(k)
            xijv=xrv+n1*ax+n2*bx+n3*cx
            yijv=yrv+n2*by+n3*cy
            zijv=zrv+n3*cz
            rij2v=xijv**2+yijv**2+zijv**2
            rijv=sqrt(rij2v)
         else 
            xijv=xij
            yijv=yij
            zijv=zij
            rij2v=rij2
            rijv=rij
         endif

         if (rij2.lt.0.003) then
            print*,'***** Short distance: ',sqrt(rij2)
            stop
         endif
         eik=0.0d0
         ei=0.0d0
         ek=0.0d0
         if (use_mpole) then
c
c     increment the multipole energy and derivative expressions
c           
            call empik1 (ii,k,-xij,-yij,-zij,rij,rij2,rpi,rpk,
     +           indi,indk,eik,ei,ek,dm,dmi,dmk,dp,dpi,dpk,utu)
            em = em + eik * factor

            if (factor.ne.1) then
               do i1 = 1,3
                  dm(i1) = dm(i1) * factor
                  do i2 = 1,3
                     dmi(i1,i2)=dmi(i1,i2) * factor
                     dmk(i1,i2)=dmk(i1,i2) * factor
                  enddo
               enddo
            endif

            dem(1,ii) = dem(1,ii) - dm(1) + dmi(1,1)
            dem(2,ii) = dem(2,ii) - dm(2) + dmi(1,2)
            dem(3,ii) = dem(3,ii) - dm(3) + dmi(1,3)
            dem(1,iz) = dem(1,iz) + dmi(2,1)
            dem(2,iz) = dem(2,iz) + dmi(2,2)
            dem(3,iz) = dem(3,iz) + dmi(2,3)
            dem(1,ix) = dem(1,ix) + dmi(3,1)
            dem(2,ix) = dem(2,ix) + dmi(3,2)
            dem(3,ix) = dem(3,ix) + dmi(3,3)
c                       if (ii.ne. jj) then
            dem(1,jj) = dem(1,jj) + is1*(dm(1) + dmk(1,1))
            dem(2,jj) = dem(2,jj) + is2*(dm(2) + dmk(1,2))
            dem(3,jj) = dem(3,jj) + is3*(dm(3) + dmk(1,3))
            dem(1,kz) = dem(1,kz) + is1*dmk(2,1)
            dem(2,kz) = dem(2,kz) + is2*dmk(2,2)
            dem(3,kz) = dem(3,kz) + is3*dmk(2,3)
            dem(1,kx) = dem(1,kx) + is1*dmk(3,1)
            dem(2,kx) = dem(2,kx) + is2*dmk(3,2)
            dem(3,kx) = dem(3,kx) + is3*dmk(3,3)
c                     endif
            dedx=-dm(1)+dmi(1,1) + dmi(2,1) + dmi(3,1)
            dedy=-dm(2)+dmi(1,2) + dmi(2,2) + dmi(3,2)
            dedz=-dm(3)+dmi(1,3) + dmi(2,3) + dmi(3,3)
            cellder(1)=cellder(1) + (n1+de1) * dedx
            cellder(2)=cellder(2) + (n2+de2) * dedy
            cellder(3)=cellder(3) + (n3+de3) * dedz
            cellder(4)=cellder(4) + (n2+de2) * dedx
            cellder(5)=cellder(5) + (n3+de3) * dedx
            cellder(6)=cellder(6) + (n3+de3) * dedy

c
c     increment the polarization energy and derivative expressions
c
            if (use_polar) then
               if (factor.ne.1) then
                  do i1 = 1,3
                     dp(i1) = dp(i1) * factor
                     do i2 = 1,3
                        dpi(i1,i2)=dpi(i1,i2) * factor
                        dpk(i1,i2)=dpk(i1,i2) * factor
                     enddo
                  enddo
               endif
                     
               ep = ep + (ei + ek) * factor
               
               dep(1,ii) = dep(1,ii) - dp(1) + dpi(1,1)
               dep(2,ii) = dep(2,ii) - dp(2) + dpi(1,2)
               dep(3,ii) = dep(3,ii) - dp(3) + dpi(1,3)
               dep(1,iz) = dep(1,iz) + dpi(2,1)
               dep(2,iz) = dep(2,iz) + dpi(2,2)
               dep(3,iz) = dep(3,iz) + dpi(2,3)
               dep(1,ix) = dep(1,ix) + dpi(3,1)
               dep(2,ix) = dep(2,ix) + dpi(3,2)
               dep(3,ix) = dep(3,ix) + dpi(3,3)
               dep(1,jj) = dep(1,jj) + is1*(dp(1) + dpk(1,1))
               dep(2,jj) = dep(2,jj) + is2*(dp(2) + dpk(1,2))
               dep(3,jj) = dep(3,jj) + is3*(dp(3) + dpk(1,3))
               dep(1,kz) = dep(1,kz) + is1*dpk(2,1)
               dep(2,kz) = dep(2,kz) + is2*dpk(2,2)
               dep(3,kz) = dep(3,kz) + is3*dpk(2,3)
               dep(1,kx) = dep(1,kx) + is1*dpk(3,1)
               dep(2,kx) = dep(2,kx) + is2*dpk(3,2)
               dep(3,kx) = dep(3,kx) + is3*dpk(3,3)
                     
               dedx= -dp(1) + dpi(1,1) + dpi(2,1) + dpi(3,1)
               dedy= -dp(2) + dpi(1,2) + dpi(2,2) + dpi(3,2)
               dedz= -dp(3) + dpi(1,3) + dpi(2,3) + dpi(3,3)
               cellder(1)=cellder(1) + (n1+de1) * dedx
               cellder(2)=cellder(2) + (n2+de2) * dedy
               cellder(3)=cellder(3) + (n3+de3) * dedz
               cellder(4)=cellder(4) + (n2+de2) * dedx
               cellder(5)=cellder(5) + (n3+de3) * dedx
               cellder(6)=cellder(6) + (n3+de3) * dedy

            endif
         endif
         
c Charge-Charge energy and derivatives
         if (use_charge) then
            if (nt14.eq.1) then
               fc=factor/chgscale
            else
               fc=factor
            endif
            
            qq=fc*pchg(ii)*pchg(jj)*electric / dielec
            e=qq/rij
            dedr=-qq/rij2
c
c     form the chain rule terms for derivative expressions
c
            dedr = dedr / rij
            dedx = dedr * xij
            dedy = dedr * yij
            dedz = dedr * zij
c
c     increment the 'multipole' energy and derivative expressions
c
            em = em + e
            dem(1,ii) = dem(1,ii) + dedx
            dem(2,ii) = dem(2,ii) + dedy
            dem(3,ii) = dem(3,ii) + dedz
            dem(1,jj) = dem(1,jj) - is1 * dedx
            dem(2,jj) = dem(2,jj) - is2 * dedy
            dem(3,jj) = dem(3,jj) - is3 * dedz
            cellder(1)=cellder(1) + (n1+de1) * dedx
            cellder(2)=cellder(2) + (n2+de2) * dedy
            cellder(3)=cellder(3) + (n3+de3) * dedz
            cellder(4)=cellder(4) + (n2+de2) * dedx
            cellder(5)=cellder(5) + (n3+de3) * dedx
            cellder(6)=cellder(6) + (n3+de3) * dedy
         endif


c        Vanderwaals energy and derivatives  
         eikv=0.0

         call evdwpair1(ii,k,jj,xijv,yijv,zijv,rij2v,rijv,
     +        rij2,rij,ddispdr,eikv,eikd,eikr,de,ani_i,ani_k,nt14)
         ev = ev + eikv * factor
         edisp = edisp + eikd * factor
         erep = erep + eikr * factor

         if (factor.ne.1.0d0) then
            ddispdr=ddispdr*factor
            do i1=1,6
               do i2=1,3
                  de(i2,i1) = de(i2,i1)*factor
               enddo
            enddo
         endif

         if (ii.eq.iv) then
            dev(1,ii) = dev(1,ii) + de(1,1)
            dev(2,ii) = dev(2,ii) + de(2,1)
            dev(3,ii) = dev(3,ii) + de(3,1)
            dedx = de(1,1)
            dedy = de(2,1)
            dedz = de(3,1)
            if (ani_i) then
               dev(1,iz) = dev(1,iz) + de(1,2)
               dev(2,iz) = dev(2,iz) + de(2,2)
               dev(3,iz) = dev(3,iz) + de(3,2)
               dev(1,ix) = dev(1,ix) + de(1,3)
               dev(2,ix) = dev(2,ix) + de(2,3)
               dev(3,ix) = dev(3,ix) + de(3,3)
               dedx=dedx + de(1,2) + de (1,3)
               dedy=dedy + de(2,2) + de (2,3)
               dedz=dedz + de(3,2) + de (3,3)
            endif
         else
            dedx = de(1,1)
            dedy = de(2,1)
            dedz = de(3,1)
            dev(1,ii) = dev(1,ii) + de(1,1) * redi
            dev(2,ii) = dev(2,ii) + de(2,1) * redi
            dev(3,ii) = dev(3,ii) + de(3,1) * redi
            dev(1,iv) = dev(1,iv) + de(1,1) * rediv
            dev(2,iv) = dev(2,iv) + de(2,1) * rediv
            dev(3,iv) = dev(3,iv) + de(3,1) * rediv
         end if
         if (jj .eq. kv) then
            dev(1,jj) = dev(1,jj) + is1 * de(1,6)
            dev(2,jj) = dev(2,jj) + is2 * de(2,6)
            dev(3,jj) = dev(3,jj) + is3 * de(3,6)
            if (ani_k) then
               dev(1,kz) = dev(1,kz) + is1*de(1,4)
               dev(2,kz) = dev(2,kz) + is2*de(2,4)
               dev(3,kz) = dev(3,kz) + is3*de(3,4)
               dev(1,kx) = dev(1,kx) + is1*de(1,5)
               dev(2,kx) = dev(2,kx) + is2*de(2,5)
               dev(3,kx) = dev(3,kx) + is3*de(3,5)
            endif
         else
            redk = kred(jj)
            redkv = 1.0d0 - redk
            dev(1,jj) = dev(1,jj) + is1 * de(1,6)*redk
            dev(2,jj) = dev(2,jj) + is2 * de(2,6)*redk
            dev(3,jj) = dev(3,jj) + is3 * de(3,6)*redk
            dev(1,kv) = dev(1,kv) + is1 * de(1,6) * redkv
            dev(2,kv) = dev(2,kv) + is2 * de(2,6) * redkv
            dev(3,kv) = dev(3,kv) + is3 * de(3,6) * redkv
         end if
c Derivatives for dispersion without reduced coordinates  
         if (ddispdr.ne.0.0d0) then
            ddispdr=ddispdr/rij
            ddispdx=ddispdr*xij
            ddispdy=ddispdr*yij
            ddispdz=ddispdr*zij
            dev(1,ii) = dev(1,ii) + ddispdx
            dev(2,ii) = dev(2,ii) + ddispdy
            dev(3,ii) = dev(3,ii) + ddispdz
            dev(1,jj) = dev(1,jj) - is1 * ddispdx
            dev(2,jj) = dev(2,jj) - is2 * ddispdy
            dev(3,jj) = dev(3,jj) - is3 * ddispdz
            dedx = dedx + ddispdx
            dedy = dedy + ddispdy
            dedz = dedz + ddispdz
         endif
         
         cellder(1)=cellder(1) + (n1+de1) * dedx
         cellder(2)=cellder(2) + (n2+de2) * dedy
         cellder(3)=cellder(3) + (n3+de3) * dedz
         cellder(4)=cellder(4) + (n2+de2) * dedx
         cellder(5)=cellder(5) + (n3+de3) * dedx
         cellder(6)=cellder(6) + (n3+de3) * dedy

      enddo
      
      enb= em + ep + ev
      if (verbose) then
         if (em .ne. 0.0) print 111,'E-multipole     ',em
         if (uinner .ne. 0.0) print 111,'incl. U2-inner  ',uinner
         if (ep .ne. 0.0) print 111,'E-polarisation  ',ep
         if (uinnerpol .ne. 0.0) print 111,'incl. U2-pol    ',uinnerpol
         if (ev .ne. 0.0) print 111,'E-Vanderwaals   ',ev
         if (etail .ne. 0.0) print 111,'incl. E-tail    ',etail
         print 111,'E-nonbe         ',enb
      endif

      do i = 1, n
         do j = 1, 3
            derivs(j,i) = derivs(j,i) + dem(j,i) + dev(j,i) + dep(j,i)
         end do
      end do
      
c Reset dielectric constant to old value for intramolecular interaction:
      dielec=dlcsave
      
      return
      end

      


