      subroutine trans(nrp,amass,x,ireftr,xcm,phi,theta,psi)

c B.P. van Eijck, 1998
C     Transforms atomic coordinates to a reference axes system.
C     Input:
C         NRP: number of atoms
C         AMASS(1..NRP): atomic masses
C         X(1..3*NRP): atomic coordinates
C         IREFTR(1..3): reference atoms

C     The reference system is chosen as follows:
C        The three atoms IREFTR(1..3) define the x,y-plane, and
C        the two atoms IREFTR(1..2) define the x-axis:
C           Calculate vectors v1=r1-r2 and v4=r3-r2
C           Find the x-axis by normalizing v1
C           Find the z-axis from v3=[v1*v4] and normalizing v3
C           Find the y-axis from v2=[v3*v1]
C        If symmetry is present, the z-axis should be the unique axis.
C     Output:
C         XCM(1..3): average coordinates (H atoms excluded)
C         XTR(1..3*NRP): coordinates in reference system
C         PHI, THETA, PSI: Euler angles (degrees) corresponding to TRMAT

C     Calls function RANDOM.
      
      implicit none
      include 'math.i'
      real*8 amass(*),x(*),xcm(*),trmat(3,3),v1(3),v2(3),v3(3),
     +     v4(3),phi,theta,psi,rsq1,rsq3,random,c,costh
      integer ireftr(*),i,nrp,i1,i2,i3,m,nz,is,n


C     Determine center of non-H atoms
      nz=0
      do m=1,3
         xcm(m)=0.0
      enddo
      do i=1,nrp
         if (amass(i).gt.3.5) then
            nz=nz+1
            do m=1,3
               xcm(m)=xcm(m)+x(3*i-3+m)
            enddo
         endif
      enddo
      do m=1,3
         xcm(m)=xcm(m)/nz
      enddo

C     Determine the axes system
      i1=ireftr(1)
      i2=ireftr(2)
      i3=ireftr(3)
      is=51297
      do m=1,3
         v1(m)=x(3*i1-3+m)-x(3*i2-3+m)
         v4(m)=x(3*i3-3+m)-x(3*i2-3+m)
      enddo
 220  continue
      v3(1)=v1(2)*v4(3)-v1(3)*v4(2)
      v3(2)=v1(3)*v4(1)-v1(1)*v4(3)
      v3(3)=v1(1)*v4(2)-v1(2)*v4(1)
      rsq1=0
      rsq3=0
      do m=1,3
         rsq1=rsq1+v1(m)**2
         rsq3=rsq3+v3(m)**2
      enddo
      if (rsq1.lt.0.00001) then
         print*,' ***** TRANS: no reference distance' ,rsq1
         write(16,*) ' ***** TRANS: no reference distance' ,rsq1
         print 10,i1,(x(3*i1-3+m),m=1,3)
         print 10,i2,(x(3*i2-3+m),m=1,3)
         print 10,i3,(x(3*i3-3+m),m=1,3)
         return
      endif
      if (rsq3.lt.0.00001) then
         print*,' ***** TRANS: atoms nearly linear:',rsq3
         write(16,*) ' ***** TRANS: atoms nearly linear:',rsq3
         print 10,i1,(x(3*i1-3+m),m=1,3)
         print 10,i2,(x(3*i2-3+m),m=1,3)
         print 10,i3,(x(3*i3-3+m),m=1,3)
         v4(1)=random(is)
         v4(2)=random(is)
         v4(3)=random(is)
         goto 220
      endif
      do m=1,3
         v1(m)=v1(m)/sqrt(rsq1)
         v3(m)=v3(m)/sqrt(rsq3)
      enddo
      v2(1)=v3(2)*v1(3)-v3(3)*v1(2)
      v2(2)=v3(3)*v1(1)-v3(1)*v1(3)
      v2(3)=v3(1)*v1(2)-v3(2)*v1(1)

C     Determine the transformation matrix
      do m=1,3
         trmat(m,1)=v1(m)
         trmat(m,2)=v2(m)
         trmat(m,3)=v3(m)
      enddo

C     Check orthogonality
c      do 280 m=1,3
c         do 270 n=1,3
c            v4(n)=0
c            do 260 k=1,3
c               v4(n)=v4(n)+trmat(n,k)*trmat(m,k)
c 260        continue
c 270     continue
c         print*,(v4(k),k=1,3)
c 280  continue

C     Transform coordinates to reference axes
c      do i=1,nrp
c         do m=1,3
c            c=0.0
c            do n=1,3
c               c=c+trmat(n,m)*(x(3*i-3+n)-xcm(n))
c            enddo
c            xtr(3*i-3+m)=c
c         enddo
c      enddo

C     Determine the Euler angles
      costh=trmat(3,3)
      if (abs(costh).gt.1) costh=sign(1.0,costh)
      theta=acos(costh)*radian
      if (abs(costh).le.0.9999) then
         psi=atan2(trmat(1,3),trmat(2,3))*radian
         phi=atan2(trmat(3,1),-trmat(3,2))*radian
      else
         phi=0.0
         psi=atan2(trmat(1,2)*costh,trmat(2,2)*costh)*radian
C        This is an arbitrary choice, only (PHI + PSI) is determinable.
      endif
      if (phi.lt.-0.1) phi=phi+360.0
      if (psi.lt.-0.1) psi=psi+360.0
 10   format(i5,3f10.5)
      end
