c
c
c
c     ##########################################################
c     ##                                                      ##
c     ##  program agda-- adiabatic Gaussian Density Annealing ##
c     ##                                                      ##
c     ##########################################################
c
c
c     "agda" implements the Adiabatic Gaussian  Density Annealing Method
c     for global optimization. This method  that implements a 
c     conjugate gradient optimization on differently annealed  
c     potential surfaces. The widths of the Gaussian densities  
c     are controlled by a numerical integrator. This can also be
c     interpreted as a variable gaussian width DEM.
c     
c
c     literature reference :
c
c     J. Ma and J. E. Straub, "Simulated Annealing using the
c     Classical Density Distribution", Journal of Chemical Physics,
c     101, 533-541 (1994)
c
c
c     C.Tsoo and C.L. Brooks III "Cluster structure determintation
c     using Gaussian density distribution global minimization 
c     methods", Journal of Chemical Physics, 101, 6405-6411. (1994)
c
c
c
      program agda
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'files.i'
      include 'iounit.i'
      include 'minima.i'
      include 'potent.i'
      include 'gauss.i'
      include 'molcul.i'
      integer maxgda
      parameter (maxgda=maxatm)
      integer i,igda,lext,freeunit
      integer iagda,nstep
      integer nvar,nok,nbad
      real*8 bstart,bstop
      real*8 random,boxsize
      real*8 eps,h1,hmin
      real*8 minimum,grdmin
      real*8 xcm,ycm,zcm
      real*8 gda1,gda2
      real*8 m2init(maxatm)
      real*8 xx(maxgda)
      character*1 answer
      character*6 mode,method
      character*7 ext,status
      character*60 gdafile
      character*80 record,string
      logical exist,randomize
      external gda1,gda2,gda3,writeout
c
c
c     initialize and get the structure to be optimized
c
      call initial
      call getxyz
c
c     set flag for GDA method, then set up molecular mechanics
c
      use_gda = .true.
      call mechanic
c
c     store the initial values of the squared mean Gaussian width
c
      do i = 1, n
         m2init(i) = m2(1)
      end do
c
c     see if random coordinates are desired as starting structures
c
      randomize = .true.
      randomize = .false.
      call nextarg (answer,exist)
      if (.not. exist) then
        write (iout,10)
   10   format (/,' Use Randomized Initial Coordinates [N] :  ',$)
        read (input,20)  answer
   20   format (a1)
      end if
      call upcase (answer)
      if (answer .eq. 'Y')  randomize = .true.
      if (randomize)  boxsize = 10.0d0 * (dble(n))**(1.0d0/3.0d0)
c
c     get initial and final values of inverse temperature
c
      bstart = -1.0d0
      bstop = -1.0d0
      call nextarg (string,exist)
      if (exist)  read (string,*,err=30,end=30)  bstart
      call nextarg (string,exist)
      if (exist)  read (string,*,err=30,end=30)  bstop
   30 continue
      if (bstart.le.0.0d0 .or. bstop.le.0.0d0) then
         write (iout,40)
   40    format (/,' Enter Initial and Final Beta [0.01, 10**10] :  ',$)
         read (input,50)  record
   50    format (a80)
         read (record,*,err=60,end=60)  bstart,bstop
   60    continue
      end if
      if (bstart .le. 0.0d0)  bstart = 0.01d0
      if (bstop .le. 0.0d0)  bstop = 1.0d10
c
c
c     get the number of search directions for local search
      neigen = 0
      call nextarg (string,exist)
      if(exist) read (string,*,err=70,end=70)  neigen
   70 continue
      if(neigen .le. 0) then
        write(iout,80) 
   80   format(/,' Enter Number of Local  Search Directions [0] :  ',$)
        read (input,90)  neigen
   90   format(i10)
      end if
      if(neigen .le. 0) neigen = 0

c
c     decide whether to generate branches
      brnch = 0
      use_branch = .false.
      if(neigen .gt. 0) then
        call nextarg (answer,exist)
        if (.not. exist) then
          write (iout,100)
  100     format(/,' Generate multiple branches during AGDA',
     &             ' [N] :  ',$)
          read (input,120)  answer
  120     format (a1)
        end if
        call upcase (answer)
        if (answer .eq. 'Y')  use_branch = .true.
      end if
c
c     write out a copy of coordinates for later update
c
      lext = 3
      call numeral (iagda,ext,lext)
      gdafile = filename(1:leng)//'.'//ext(1:lext)
      call version (gdafile,'new')
      igda = freeunit ()
      open (unit=igda,file=gdafile,status='new')
      call prtxyz (igda)
      close (unit=igda)
      outfile = gdafile
c     
c     set an initial box size and generate random coordinates
c     
      if (randomize) then
        do i = 1, n
          x(i) = boxsize * random ()
          y(i) = boxsize * random ()
          z(i) = boxsize * random ()
        end do
      end if
c     
c     translate M2's to parameters for numerical integration
c     
      nvar = 0
      do i = 1, n
        nvar = nvar + 1
        xx(nvar) = m2init(i)
      end do
c     
c     make changes to the potential to use potential smoothing
c     
      use_gda = .true.
      use_geom = .true.
      use_lj = .false.
      use_gauss = .true.
c     
c     make the call to the Bulirsch-Stoer integration routine
c     
      nstep = 0
      status = '       '
      eps = 1.0d-8
      h1 = 0.1d0
      hmin = 0.0d0
      write (iout,130)
  130 format (//,' Adiabatic Gaussian Density Annealing   :',
     &        //,' BS Step',5x,'Kelvin   ',6x,'Energy',
     &        9x,'Rg',8x,'Log(M2)',7x,'Status',/)
      varlast = 0.0d0
      path ='main  '
      beta = bstart
      bfinal = bstop
      call gdastat (nstep,bstart,xx,status)
      call diffeq (nvar,xx,bstart,bstop,eps,h1,hmin,nok,nbad,gda1)
      nstep = nok + nbad
c     
c     make changes to the potential for standard optimization
c     
      use_gda = .false.
      use_geom = .false.
      use_lj = .true.
      use_gauss = .false.
c     
c     make the call to the energy minimization routine
c     
      nvar = 0
      do i = 1,n
        nvar = nvar + 1
        xx(nvar) = x(i)
        nvar = nvar + 1
        xx(nvar) = y(i)
        nvar = nvar + 1
        xx(nvar) = z(i)
      end do
      mode = 'dtncg '
      method = 'auto  '
      grdmin = 0.0001d0
      nextiter = nstep + 1
      call tncg (mode,method,nvar,xx,minimum,grdmin,
     &  gda2,gda3,writeout)
      write (iout,110)  

  110 format (/,' AGDA: Global Energy Minimum :',f15.4)
c     
c     translate optimization parameters into atomic coordinates
c     
      nvar = 0
      do i = 1, n
        nvar = nvar + 1
        x(i) = xx(nvar)
        nvar = nvar + 1
        y(i) = xx(nvar)
        nvar = nvar + 1
        z(i) = xx(nvar)
      end do
c     
c     move the center of mass to the origin
c     
      xcm = 0.0d0
      ycm = 0.0d0
      zcm = 0.0d0
      do i = 1, n
        xcm = xcm + x(i)
        ycm = ycm + y(i)
        zcm = zcm + z(i)
      end do
      xcm = xcm / dble(n)
      ycm = ycm / dble(n)
      zcm = zcm / dble(n)
      do i = 1, n
        x(i) = x(i) - xcm
        y(i) = y(i) - xcm
        z(i) = z(i) - xcm
      end do
c     
c     write the final coordinates into a file
c     
      igda = freeunit ()
      open (unit=igda,file=gdafile,status='old')
      rewind (unit=igda)
      call prtxyz (igda)
      close (unit=igda)
      end
c
c
c     ##################################################################
c     ##                                                              ##
c     ##  function gda1  --  gradient for gaussian density annealing  ##
c     ##                                                              ##
c     ##################################################################
c
c
      function gda1 (bval,xx,g)
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'iounit.i'
      include 'gauss.i'
      integer maxgda
      parameter (maxgda=maxatm)
      integer i,nvar
      integer hinit(maxvar),hstop(maxvar)
      integer hindex(maxhess)
      real*8 energy,bval
      real*8 gda1,sum
      real*8 hdiag(maxvar),h(maxhess)
      real*8 xx(maxgda),g(maxgda)
c
c
      beta = bval
c     translate optimization parameters to M2's
c
      nvar = 0
      do i = 1, n
        nvar = nvar + 1
        m2(i) = xx(nvar)
        if (m2(i) .lt. 0.0d0) then
          m2(i) = -m2(i)
        end if
      end do
c
c     compute and store the energy and gradient
c
      gda1 = energy ()
c
c
c     compute and store the Hessian elements
c
      call hessian (h,hinit,hstop,hindex,hdiag)
c
c     translate the Hessian diagonal into a dM2/dbeta vector
c
      nvar = 0
      do i = 1, n
         nvar = nvar + 1
         sum = hdiag(3*i-2) + hdiag(3*i-1) + hdiag(3*i)
         g(nvar) = -(m2(i)/3.0d0)**2 * sum 
      end do
      return
      end
c
c
c     ################################################################
c     ##                                                            ##
c     ##  function gda2  --  energy/gradient for TNCG optimization  ##
c     ##                                                            ##
c     ################################################################
c
c
      function gda2 (xx,g)
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      integer i,nvar
      real*8 gda2,e,derivs(3,maxatm)
      real*8 xx(maxvar),g(maxvar)
c
c
c     translate optimization parameters to atomic coordinates
c
      nvar = 0
      do i = 1, n
         nvar = nvar + 1
         x(i) = xx(nvar)
         nvar = nvar + 1
         y(i) = xx(nvar)
         nvar = nvar + 1
         z(i) = xx(nvar)
      end do
c
c     compute and store the energy and gradient
c
      call gradient (e,derivs)
      gda2 = e
c
c     store atom gradients as optimization gradient, also
c     translate the coordinates of each active atom; the
c     latter may be needed when using periodic boundaries
c
      nvar = 0
      do i = 1, n
         nvar = nvar + 1
         xx(nvar) = x(i)
         g(nvar) = derivs(1,i)
         nvar = nvar + 1
         xx(nvar) = y(i)
         g(nvar) = derivs(2,i)
         nvar = nvar + 1
         xx(nvar) = z(i)
         g(nvar) = derivs(3,i)
      end do
      return
      end
c
c
c     #################################################################
c     ##                                                             ##
c     ##  subroutine gda3  --  Hessian values for TNCG optimization  ##
c     ##                                                             ##
c     #################################################################
c
c
      subroutine gda3 (mode,xx,h,hinit,hstop,hindex,hdiag)
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      integer i,nvar
      integer hinit(maxvar),hstop(maxvar)
      integer hindex(maxhess)
      real*8 xx(maxvar),hdiag(maxvar),h(maxhess)
      character*4 mode
c
c
c     translate optimization parameters to atomic coordinates
c
      if (mode .eq. 'none')  return
      nvar = 0
      do i = 1, n
         nvar = nvar + 1
         x(i) = xx(nvar)
         nvar = nvar + 1
         y(i) = xx(nvar)
         nvar = nvar + 1
         z(i) = xx(nvar)
      end do
c
c     compute and store the Hessian elements
c
      call hessian (h,hinit,hstop,hindex,hdiag)
c
c     translate the coordinates of each active atom;
c     this may be needed when using periodic boundaries
c
      nvar = 0
      do i = 1, n
         nvar = nvar + 1
         xx(nvar) = x(i)
         nvar = nvar + 1
         xx(nvar) = y(i)
         nvar = nvar + 1
         xx(nvar) = z(i)
      end do
      return
      end
c
c
c
c     ###############################################################
c     ##                                                           ##
c     ##  subroutine gdastat  --  compute GDA trajectory averages  ##
c     ##                                                           ##
c     ###############################################################
c
c
      subroutine gdastat (nstep,bval,xx,status)
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'iounit.i'
      include 'math.i'
      include 'gauss.i'
      include 'minima.i'
      include 'inform.i'
      include 'units.i'
      integer maxgda,oldprt
      parameter (maxgda=maxatm)
      integer i,nstep,nvar,nmin
      real*8 xx(maxgda),xc(maxvar)
      real*8 e,energy,rg,m2ave,temp
      real*8 grdmin,minimum,bval
      real*8 mean,std,eps
      logical oldverb
      logical use_local
      character*6 mode,method
      character*7 status
      external gda2,gda3,writeout
c
c
c     translate optimization parameters to M2's
c
      beta = bval
      nvar = 0
      do i = 1, n
        nvar = nvar + 1
        m2(i) = abs(xx(nvar))
      end do
c
c     translate coordinates to optimization parameters
      nmin = 0
      do i = 1, n
        nmin = nmin + 1
        xc(nmin) = x(i) 
        nmin = nmin + 1
        xc(nmin) = y(i)
        nmin = nmin + 1
        xc(nmin) = z(i)
      end do
c
c
c     make the call to the energy minimization routine
c
      mode = 'dtncg '
      method = 'auto  '
      grdmin = 0.0001d0
      oldverb = verbose
      oldprt = iprint
      verbose = .false.
      iprint = 0
      call tncg (mode,method,nmin,xc,minimum,grdmin,
     &           gda2,gda3,writeout)     

c     translate optimization parameters to coordinates
      nmin = 0
      do i = 1, n
        nmin = nmin + 1
        x(i) = xc(nmin)
        nmin = nmin + 1
        y(i) = xc(nmin)
        nmin = nmin + 1
        z(i) = xc(nmin)
      end do
c
      if(path .eq. 'main  ') then
        call stat(n,xx,mean,std)
        temp = 1.0d0/(gasconst*beta)
        use_local = .false.
        varcrnt = beta*std
        eps = 1.0d-4
        if(varcrnt .ge. varlast+eps) then
          use_local = .true.
        end if
        
        varlast = varcrnt
        if(use_local) call loclsrch (minimum)      
c       
        e = energy ()
        
        call gyrate (rg)
        m2ave = 0.0d0
        do i = 1, n
          m2ave = m2ave + m2(i)
        end do
        m2ave = m2ave / dble(n)
c       
c       
        write (iout,10)  nstep,temp,e,rg,
     &                   log(m2ave)/logten,status
   10   format (i6,2x,4f13.4,6x,a7)
c
c       else current calculation is for a branch no output
      end if

c     save the current coordinates to a disk file
      nmin = 0
      do i = 1, n
        nmin = nmin + 1
        xc(nmin)=x(i) 
        nmin = nmin + 1
        xc(nmin)=y(i) 
        nmin = nmin + 1
        xc(nmin)=z(i)  
      end do
      call writeout (xc,nstep)
     
      return
      end
c
c
c
c     ###############################################################
c     ##                                                           ##
c     ##  subroutine loclsrch  --  local search on smooth surfaces ##
c     ##                                                           ##
c     ###############################################################
c
c
      subroutine loclsrch (minimum)
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'iounit.i'
      include 'omega.i'
      include 'refer.i'
      include 'gauss.i'
      integer i,j,k,nlocl,nsearch
      real*8 minimum,minref,minbest
      real*8 eps,rms,step(3,maxvib)
      real*8 eigen(maxvib),vects(maxvib,maxvib)
      real*8 xbest(maxatm),ybest(maxatm),zbest(maxatm)
      real*8 nu0,h0
      logical done
c
c
c     store the current coordinates as the reference set
c
      call makeref
c     
c     set parameters related to the local search procedure
c     
      done = .false.
      eps = 1.0d-4
      minref = minimum
      minbest = minimum
      nlocl = 0
c     
c     find local minimum along each of the steepest directions
c     
      dowhile (.not. done)
        nlocl = nlocl + 1
        write (iout,10)  nlocl,minref
   10   format (/,' Local Search :',10x,'Iteration',i4,
     &          5x,'Energy',f12.4,/)
        call eigencart (eigen,vects)
c       
c       
        call vibfreq(nu0)
        h0 = minref
        
c       search both directions along each eigenvector in turn
c       
        nsearch = 0
        do i = 1, neigen
          do k = 1, n
            j = 3*(k-1)
            step(1,k) = vects(j+1,n-i+1)
            step(2,k) = vects(j+2,n-i+1)
            step(3,k) = vects(j+3,n-i+1)
          end do
          nsearch = nsearch + 1
          call getref
          call climber (nsearch,minimum,step,nu0,h0)
          if (minimum .lt. minbest) then
            minbest = minimum
            do k = 1, n
              xbest(k) = x(k)
              ybest(k) = y(k)
              zbest(k) = z(k)
            end do
          end if
          do k = 1, n
            j = 3*(k-1)
            step(1,k) = -vects(j+1,n-i+1)
            step(2,k) = -vects(j+2,n-i+1)
            step(3,k) = -vects(j+3,n-i+1)
          end do
          nsearch = nsearch + 1
          call getref
          call climber (nsearch,minimum,step,nu0,h0)
          if (minimum .lt. minbest) then
            minbest = minimum
            do k = 1, n
              xbest(k) = x(k)
              ybest(k) = y(k)
              zbest(k) = z(k)
            end do
          end if
        end do
c       
c       check for convergence of the local search procedure
c       
        if (minbest .lt. minref-eps) then
          done = .false.
          minref = minbest
          call impose (n,xref,yref,zref,n,xbest,ybest,zbest,rms)
          do k = 1, n
            x(k) = xbest(k)
            y(k) = ybest(k)
            z(k) = zbest(k)
          end do
          call makeref
        else
          done = .true.
          minimum = minref
          call getref
          path = 'main  '
        end if
      end do
      return
      end
c
c
c
c
c     ###############################################################
c     ##                                                           ##
c     ##  subroutine eigencart  --  Cartesian Hessian eigenvectors ##
c     ##                                                           ##
c     ###############################################################
c
c
      subroutine eigencart(eigen,vects)
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'hescut.i'
      include 'iounit.i'
      include 'math.i'
      integer i,j,k,ihess
      integer nfreq,ndummy,nvib
      integer hinit(3,maxatm),hstop(3,maxatm),hindex(maxhess)
      real*8 h(maxhess),hdiag(3,maxatm)
      real*8 matrix((maxvib+1)*maxvib/2)
      real*8 eigen(maxvib),vects(maxvib,maxvib)
      real*8 a(maxvib+1),b(maxvib+1),p(maxvib+1),w(maxvib+1)
      real*8 ta(maxvib+1),tb(maxvib+1),ty(maxvib+1)
c
      nfreq = 3 * n
      ndummy = 0
      nvib = nfreq - 3*ndummy
c     calculate the Hessian matrix of second derivatives
c
      hesscut = 0.0d0
      call hessian (h,hinit,hstop,hindex,hdiag)
c
c     store upper triangle of the Hessian in "matrix"
c
      ihess = 0
      do i = 1, n
         do j = 1, 3
            ihess = ihess + 1
            matrix(ihess) = hdiag(j,i)
            do k = hinit(j,i), hstop(j,i)
               ihess = ihess + 1
               matrix(ihess) = h(k)
            end do
         end do
      end do
c
c     perform diagonalization to get Hessian eigenvalues
c
      call diagq (nfreq,maxvib,nfreq,matrix,eigen,vects,
     &                     a,b,p,w,ta,tb,ty)
      return
      end
c
c
c
c
c     #################################################################
c     ##                                                             ##
c     ##  subroutine climber  --  find minimum along torsional mode  ##
c     ##                                                             ##
c     #################################################################
c
c
      subroutine climber (nsearch,minimum,step,nu0,h0)
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'refer.i'
      include 'iounit.i'
      include 'math.i'
      include 'gauss.i'
      include 'potent.i'
      integer maxstep
      parameter (maxstep=500)
      integer maxgda
      parameter (maxgda=maxatm)
      integer i,kstep,nstep,nsearch
      real*8 minimum,minref,big,energy
      real*8 step(3,maxvib),estep(0:maxstep)
      real*8 nu0,h0,nu
      real*8 ddg,enthlp,entrp,prob
      real*8 xmain(maxatm),ymain(maxatm),zmain(maxatm)
      real*8 m2ref(maxgda),bref
      logical done
c
c
c     convert current reference coordinates to a Z-matrix
c
      call getref
c
c     set the maximum number of steps and the step size
c
      done = .false.
      big = 1.0d8
      minimum = big
      kstep = 0
      nstep = 65
c
c     scan the search direction for a minimization candidate
c
      dowhile (.not. done)
         if (kstep .ne. 0) then
           do i = 1, n
             x(i) = x(i) + step(1,i)
             y(i) = y(i) + step(2,i)
             z(i) = z(i) + step(3,i)
           end do
         end if
         estep(kstep) = energy ()
         if (kstep .ge. 2) then
           if (estep(kstep) .lt. estep(kstep-2) .and.
     &         estep(kstep-1) .lt. estep(kstep-2)) then
               done = .true.
               do i = 1, n
                 x(i) = x(i) - step(1,i)
                 y(i) = y(i) - step(2,i)
                 z(i) = z(i) - step(3,i)
               end do
               call loclmin (minimum)
               call vibfreq(nu) 
               enthlp = minimum - h0
               entrp  = (1.0d0/beta)*log(nu0/nu)
               if(abs(enthlp) .lt. 1.0d-4) then 
                 ddg = 0.0d0
               else
                 ddg = enthlp - entrp
               end if
               prob = 1.0d0/exp(-beta*ddg)
               if (minimum .ge. -big) then
                  write (iout,10)  nsearch,kstep+1,minimum,prob
   10             format (4x,'Search Direction',i4,10x,'Step',
     &                       i4,7x,2f12.4)
c
c                 check to see if condition for branching is satisfied
                  if(prob .lt. 1.0d0 .and. use_branch) then
                    do i = 1,n
                      xmain(i) = x(i)
                      ymain(i) = y(i)
                      zmain(i) = z(i)
                      m2ref(i) = m2(i)
                    end do
                    bref = beta
                    minref = minimum
c
c                   call routine to follow branch down to undeformed PES
c
                    path = 'branch'
                    call brnchmn
c
c                   revert back to parameters for main path
                    do i = 1,n
                      x(i)  = xmain(i)
                      y(i)  = ymain(i)
                      z(i)  = zmain(i)
                      m2(i) = m2ref(i)
                    end do
                    beta = bref
                    minimum = minref
                    path = 'main  '
                    use_gda = .true.
                    use_geom = .true.
                    use_lj = .false.
                    use_gauss = .true.
                  end if
               else
                  minimum = big
                  write (iout,20)  nsearch
   20             format (4x,'Search Direction',i4,35x,'------')
               end if
            end if
         end if
         if (kstep.ge.nstep .and. .not.done) then
            done = .true.
            write (iout,30)  nsearch
   30       format (4x,'Search Direction',i4,35x,'------')
         end if
         kstep = kstep + 1
      end do
      return
      end
c
c
c     #############################################################
c     ##                                                         ##
c     ##  subroutine loclmin  --  optimization for local search  ##
c     ##                                                         ##
c     #############################################################
c
c
c     "loclmin" performs an energy minimization in Cartesian
c     coordinate space using a limited memory quasi-Newton method
c
c
      subroutine loclmin (minimum)
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'inform.i'
      include 'minima.i'
      integer i,nvar,oldprt
      real*8 minimum,grdmin
      real*8 gda2,xx(maxvar)
      logical oldverb
      external gda2,gda3,writeout
c
c
c     translate the coordinates of each atom
c
      nvar = 0
      do i = 1, n
         nvar = nvar + 1
         xx(nvar) = x(i)
         nvar = nvar + 1
         xx(nvar) = y(i)
         nvar = nvar + 1
         xx(nvar) = z(i)
      end do
c
c     make the call to the optimization routine
c
      grdmin = 0.0001d0
      oldverb = verbose
      oldprt = iprint
      verbose = .false.
      iprint = 0
      call lmqn (nvar,xx,minimum,grdmin,gda2,writeout)
c
c     untranslate the final coordinates for each atom
c
      nvar = 0
      do i = 1, n
         nvar = nvar + 1
         x(i) = xx(nvar)
         nvar = nvar + 1
         y(i) = xx(nvar)
         nvar = nvar + 1
         z(i) = xx(nvar)
      end do
      return
      end
c
c
c
c
c
c     ##################################################################
c     ##                                                              ##
c     ##  subroutine vibfreq  -- vibrational analysis and normal modes##
c     ##                                                              ##
c     ##################################################################
c
c
c     "vibfreq" performs a vibrational normal mode analysis; the
c     Hessian matrix of second derivatives is determined and then
c     diagonalized after mass weighting; output is the product of 
c     the normal mode frequencies
c
c
      subroutine vibfreq(product)
      implicit none
      include 'sizes.i'
      include 'atmtyp.i'
      include 'atoms.i'
      include 'files.i'
      include 'hescut.i'
      include 'iounit.i'
      include 'math.i'
      include 'units.i'
      include 'warp.i'
      integer i,j,k,m,ihess
      integer nfreq,ndummy,nvib
      integer hinit(3,maxatm),hstop(3,maxatm),hindex(maxhess)
      real*8 h(maxhess),hdiag(3,maxatm)
      real*8 mass2(maxatm),factor
      real*8 matrix((maxvib+1)*maxvib/2)
      real*8 eigen(maxvib),vects(maxvib,maxvib)
      real*8 a(maxvib+1),b(maxvib+1),p(maxvib+1),w(maxvib+1)
      real*8 ta(maxvib+1),tb(maxvib+1),ty(maxvib+1)
      real*8 product
      character*1 eigunits(maxvib)
c
c
c     set up the structure and mechanics calculation
c
c     initialize various things needed for vibrations
c
      nfreq = 3 * n
      ndummy = 0
      if (nfreq.gt.maxvib .or. nfreq**2.gt.maxhess) then
         write (iout,10)
   10    format (/,4x,' VIBFREQ  --  Too many Atoms in the Molecule')
         call fatal
      end if
      do i = 1, n
        if (atomic(i) .eq. 0) then
          ndummy = ndummy + 1
          mass(i) = 0.001d0
        end if
        mass2(i) = sqrt(mass(i))
      end do
      nvib = nfreq - 3*ndummy
c
c     
c     calculate the Hessian matrix of second derivatives
c
      hesscut = 0.0d0
      call hessian (h,hinit,hstop,hindex,hdiag)
c
c     store upper triangle of the mass-weighted Hessian matrix
c
      ihess = 0
      do i = 1, n
        do j = 1, 3
          ihess = ihess + 1
          matrix(ihess) = hdiag(j,i) / mass(i)
          do k = hinit(j,i), hstop(j,i)
            m = (hindex(k)+2) / 3
            ihess = ihess + 1
            matrix(ihess) = h(k) / (mass2(i)*mass2(m))
          end do
        end do
      end do
c
c     diagonalize to get vibrational frequencies and normal modes
c     
      call diagq (nfreq,maxvib,nfreq,matrix,eigen,vects,
     &                     a,b,p,w,ta,tb,ty)
      factor = 4.6587d5 / (4.0d0 * pi**2)
      do i = 1, nvib
        if (eigen(i) .lt. 0) then
          eigunits(i) = 'I'
        else
          eigunits(i) = ' '
        end if
c        eigen(i) = sqrt(abs(factor*eigen(i)))
        eigen(i) = sqrt(abs(eigen(i)))
      end do
c     
c     compute products of vibrational frequencies
      product = 1.0d0
      do i = 7,nvib
        product = product * eigen(i)
      end do
      end
c
c
c
c     ##############################################
c     ##                                          ##
c     ##  subroutine brnchmn -- agda for a branch ##
c     ##                                          ##
c     ##############################################
c
c

c     "brnchmin" takes the current m2 values and the Cartesian coordinates
c     for a chosen branch to follows an AGDA algorithm down to the 
c     undeformed surface
c
      subroutine brnchmn
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'atmtyp.i'
      include 'files.i'
      include 'iounit.i'
      include 'minima.i'
      include 'potent.i'
      include 'gauss.i'
      integer maxgda
      parameter (maxgda=maxatm)
      integer i,ibrnch,size,freeunit
      integer nstep,nvar,nok,nbad
      real*8 bstart,bstop
      real*8 eps,h1,hmin
      real*8 minimum,grdmin
      real*8 xcm,ycm,zcm
      real*8 gda1,gda2
      real*8 xx(maxgda)
      real*8 total
      character*6 mode,method
      character*7 status,string
      character*60 brnchfile
      external gda1,gda2,gda3,writeout
c
c
c
c
      bstart = beta
      bstop = bfinal
      path = 'branch'
      brnch = brnch + 1
c
      write (iout,10)  brnch,beta
   10   format (/,4x,' AGDA Branch:',10x,'Number',i4,
     &          5x,'Beta',f12.4,/)
c     translate M2's to parameters for numerical integration
c     
      nvar = 0
      do i = 1, n
        nvar = nvar + 1
        xx(nvar) = m2(i)
      end do
c
c     make changes to the potential to use potential smoothing
c
      use_gda = .true.
      use_geom = .true.
      use_lj = .false.
      use_gauss = .true.
c
c     make the call to the Bulirsch-Stoer integration routine
c
      nstep = 0
      status = '       '
      eps = 1.0d-8
      h1 = 0.1d0
      hmin = 0.0d0
      call gdastat (nstep,bstart,xx,status)
      call diffeq (nvar,xx,bstart,bstop,eps,h1,hmin,nok,nbad,gda1)
      nstep = nok + nbad
c
c     make changes to the potential for standard optimization
c     
      use_gda = .false.
      use_geom = .false.
      use_lj = .true.
      use_gauss = .false.
c     
c     make the call to the energy minimization routine
c     
      nvar = 0
      do i = 1,n
        nvar = nvar + 1
        xx(nvar) = x(i)
        nvar = nvar + 1
        xx(nvar) = y(i)
        nvar = nvar + 1
        xx(nvar) = z(i)
      end do
      mode = 'dtncg '
      method = 'auto  '
      grdmin = 0.0001d0
      nextiter = nstep + 1
      call tncg (mode,method,nvar,xx,minimum,grdmin,
     &           gda2,gda3,writeout)
      write (iout,20)  brnch,minimum
   20 format (/,4x,' Minimum Energy for Branch ',i4,' :',f15.4)
c     
c     translate optimization parameters into atomic coordinates
c     
      nvar = 0
      do i = 1, n
        nvar = nvar + 1
        x(i) = xx(nvar)
        nvar = nvar + 1
        y(i) = xx(nvar)
        nvar = nvar + 1
        z(i) = xx(nvar)
      end do
c
c     move the center of mass to the origin
c     
      xcm = 0.0d0
      ycm = 0.0d0
      zcm = 0.0d0
      do i = 1, n
        total = total + mass(i)
        xcm = xcm + x(i)*mass(i)
        ycm = ycm + y(i)*mass(i)
        zcm = zcm + z(i)*mass(i)
      end do
      xcm = xcm / total
      ycm = ycm / total
      zcm = zcm / total
      do i = 1, n
        x(i) = x(i) - xcm
        y(i) = y(i) - xcm
        z(i) = z(i) - xcm
      end do
c     
      size = 3
      call numeral(brnch,string,size)
      brnchfile = 'branch'//'.'//string(1:size)
      ibrnch = freeunit ()
      open(unit=ibrnch,file=brnchfile,status='new')
      rewind(unit=ibrnch)
      call prtxyz(ibrnch)
      close(unit=ibrnch)
      end
