c
c
c     ###############################################################
c     ##                                                           ##
c     ##  subroutine esolv5 - smooth solvation energy & gradients  ##
c     ##                                                           ##
c     ###############################################################
c
c
c     "esolv5" calculates the macroscopic solvation energy and
c     first derivatives with respect to Cartesian coordinates
c     using the Macromodel GB/SA solvation model
c
c
      subroutine esolv5
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'deriv.i'
      include 'energi.i'
      include 'potent.i'
      include 'solute.i'
      integer i,j
      real*8 probe,aarea(maxatm)
c     
c     zero out the macroscopic solvation energy and derivatives
c     
      es = 0.0d0
      do i = 1, n
        des(1,i) = 0.0d0
        des(2,i) = 0.0d0
        des(3,i) = 0.0d0
      end do
c     
c     compute the surface area-based solvation energy term
c     
      probe = 1.4d0
      call surface (earea,aarea,darea,rsolv,vsolv,probe)
      es = es + earea
      do i = 1, n
        do j = 1, 3
          des(j,i) = des(j,i) + darea(j,i)
        end do
      end do
c     
c     get the generalized Born term for GB/SA solvation
c     
      call egborn5 (epol,depol)
      es = es + epol
      do i = 1, n
	do j = 1, 3
          des(j,i) = des(j,i) + depol(j,i)
        end do
      end do
      return
      end
c     
c
c     ###########################################################
c     ##                                                       ##
c     ##  subroutine egborn5  --  smooth Born energy for GB/SA ##
c     ##                                                       ##
c     ###########################################################
c
c
c     "egborn5" calculates the smoothed generalized Born energy and 
c     first derivatives with respect to Cartesian coordinates for 
c     the Macromodel GB/SA solvation model
c
c
      subroutine egborn5 (epol,depol)
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'math.i'
      include 'bound.i'
      include 'charge.i'
      include 'shunt.i'
      include 'solute.i'
      include 'units.i'
      include 'usage.i'
      include 'warp.i'
      integer i,k,ii,kk
      real*8 epol,e,f,fi,fik,rb,rb2
      real*8 xi,yi,zi,xr,yr,zr
      real*8 dedx,dedy,dedz,de,dwater
      real*8 r,r2,expterm,expterm2
      real*8 s1,s2,s3,s4,asqpt,erfterm
      real*8 def1,def2,def3,def4,erf
      real*8 dgb1,dgb2,dgb11,dgb12
      real*8 fgb1,fgb2,irb,irb2
      real*8 arg1,arg2,arg3,arg4
      real*8 depol(3,maxatm)
      logical iuse
c
c
c     zero out the GB/SA polarization energy and derivatives
c
      epol = 0.0d0
      do i = 1, n
        depol(1,i) = 0.0d0
        depol(2,i) = 0.0d0
        depol(3,i) = 0.0d0
      end do
      if (nion .eq. 0)  return
c     
c     set the dielectric constant and energy conversion factor
c
      dwater = 78.3d0
      f = -electric * (1.0d0 - 1.0d0/dwater)
c
c     set cutoff distances and switching function coefficients
c
      call switch ('CHARGE')
c
c     calculate GB/SA electrostatic polarization energy term
c
      do ii = 1, nion
        i = iion(ii)
        iuse = use(i)
        xi = x(i)
        yi = y(i)
        zi = z(i)
        fi = f * pchg(ii)
        do kk = ii, nion
          k = iion(kk)
          if (iuse .or. use(k)) then
            xr = xi - x(k)
            yr = yi - y(k)
            zr = zi - z(k)
            r2 = xr*xr + yr*yr + zr*zr
            r = sqrt(r2)
            fik = fi * pchg(kk)
            rb2 = rborn(i) * rborn(k)
            rb = sqrt(rb2)
            if (deform .ne. 0.0d0) then
              s1 = sqrt(r2 + 1.25d0*rb2)
              s2 = s1*s1
              s3 = s2*s1
              s4 = sqra*rb
              arg1 = s1 + s4
              arg2 = s1 - s4
              arg3 = (arg1*arg1)
              arg4 = (arg2*arg2) 
              asqpt = rb2 + deform
              def1 = 0.5d0/sqrt(deform)
              def2 = (1.0d0/sqrt(asqpt))**3
              def3 = (1.0d0/sqrt(asqpt))**5
              def4 = def1*def1
              erfterm = erf(arg1*def1) + erf(arg2*def1)
              expterm = exp(-0.25d0*r2/asqpt)
	        expterm2 = exp(-arg3*def4) + exp(-arg4*def4)
              fgb1 = 0.5d0*erfterm/s1
	        if (i .eq. k) fgb1 = 2.0d0*fgb1
              fgb2 = screen*rb2*def2*expterm
              dgb11 = -0.5d0*r*erfterm/s3
              dgb12 = 0.5d0*r*expterm2 /(s2*sqrt(pi*deform)) 
              dgb1 = dgb11 + dgb12
              dgb2 = -0.5d0*screen*r*rb2*expterm*def3
              de = fik * (dgb1 + dgb2)
            else
              irb = screen/rb
              irb2 = 0.5d0/rb2
              fgb1 = 1.0d0/sqrt(r2 + 1.25d0*rb2)
              expterm = exp(-0.25d0*r2/rb2)
              fgb2 = irb*expterm
              dgb1 = -fgb1**3
              dgb2 = -irb2 * fgb2
              de = fik * r*(dgb1 + dgb2)
            end if
c           
            e = fik *(fgb1 + fgb2)
c           increment the overall energy and derivative expressions
c           
            if(i .eq. k) then
              e = 0.5d0*e
            else
              dedx = de * xr / r
              dedy = de * yr / r
              dedz = de * zr / r
              depol(1,i) = depol(1,i) + dedx
              depol(2,i) = depol(2,i) + dedy
              depol(3,i) = depol(3,i) + dedz
              depol(1,k) = depol(1,k) - dedx
              depol(2,k) = depol(2,k) - dedy
              depol(3,k) = depol(3,k) - dedz
c             
            end if
            epol = epol + e
          end if
        end do
      end do
c
      return
      end
