c
c
c
c     ##################################################################
c     ##                                                              ##
c     ##  subroutine esolv7  --  smooth macroscopic solvation energy  ##
c     ##                                                              ##
c     ##################################################################
c
c
c     "esolv7" calculates the smoothed macroscopic solvation energy 
c     using an approximation to the Macromodel GB/SA solvation model
c
c
      subroutine esolv7
      implicit none
      include 'sizes.i'
      include 'action.i'
      include 'analyz.i'
      include 'atmtyp.i'
      include 'atoms.i'
      include 'energi.i'
      include 'iounit.i'
      include 'inform.i'
      include 'potent.i'
      include 'solute.i'
      integer i   
      real*8 probe,darea(3,maxatm)
      logical header,huge

c
c
c     zero out the macroscopic solvation energy
c
      nes = 0
      es = 0.0d0
      do i = 1, n
        aes(i) = 0.0d0
      end do
c
c     compute the surface area-based solvation energy term
c
      header = .true.
      probe = 1.4d0
      call surface (earea,aarea,darea,rsolv,vsolv,probe)
      nes = nes + n
      es = es + earea
      do i = 1, n
         aes(i) = aes(i) + aarea(i)
      end do
c
c     get the generalized Born term for GB/SA solvation
c
      call egborn7
      es = es + epol
      do i = 1, n
        aes(i) = aes(i) + aepol(i)
      end do
c     print a warning if the energy of any atom is large
c
      do i = 1, n
        huge = (abs(aes(i)) .gt. 25.0d0)
        if (debug .or. (verbose.and.huge)) then
          if (header) then
            header = .false.
            write (iout,10)
   10       format (/,' Individual Atomic Solvation Energy',
     &                ' Terms :',
     &               //,' Type',11x,'Atom Name',42x,'Energy',/)
          end if
          write (iout,20)  i,name(i),aes(i)
   20     format (' Solvate',7x,i5,'-',a3,37x,f12.4)
        end if
      end do
      return
      end
c
c
c     ################################################################
c     ##                                                            ##
c     ##  subroutine egborn7  -- smoothed polarization / Born energy##
c     ##                                                            ##
c     ################################################################
c
c
c     "egborn7" calculates the generalized Born energy term
c     using an approximation to the Macromodel GB/SA solvation model
c
c
      subroutine egborn7
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'action.i'
      include 'analyz.i'
      include 'bound.i'
      include 'charge.i'
      include 'energi.i'
      include 'shunt.i'
      include 'solute.i'
      include 'units.i'
      include 'usage.i'
      include 'warp.i'
      integer i,k,ii,kk
      real*8 e,f,fi,fik,fgb1,fgb2
      real*8 rb,irb,rb2,r2
      real*8 xi,yi,zi,xr,yr,zr
      real*8 dwater,def1,def2
      real*8 s1,s2,asqpt
      real*8 erf,erfterm,expterm
      real*8 arg1,arg2
      logical iuse
c
c
c     zero out the GB/SA polarization energy
c
      epol = 0.0d0
      if (nion .eq. 0)  return
c     
c     set the dielectric constant and energy conversion factor
c     
      dwater = 78.3d0
      f = -electric * (1.0d0 - 1.0d0/dwater)
c     
c     set cutoff distances and switching function coefficients
c     
      call switch ('CHARGE')
c     
c
c     calculate GB/SA electrostatic polarization energy term
c     
      do ii = 1, nion
        i = iion(ii)
        iuse = use(i)
        xi = x(i)
        yi = y(i)
        zi = z(i)
        fi = f * pchg(ii)
        do kk = ii, nion
          k = iion(kk)
          if (iuse .or. use(k)) then
            xr = xi - x(k)
            yr = yi - y(k)
            zr = zi - z(k)
            r2 = xr*xr + yr*yr + zr*zr
            fik = fi * pchg(kk)
            rb2 = rborn(i) * rborn(k)
            rb = sqrt(rb2)
            irb = screen/ rb
            if(deform .ne. 0.0d0) then
              s1 =   sqrt(r2 + 1.25d0*rb2)
              s2 =   sqra*rb
              arg1 = s1 + s2
              arg2 = s1 - s2
              asqpt = rb2 + deform	 
              def1 = 0.5d0/sqrt(deform)
              def2 = (1.0d0/sqrt(asqpt))**3
              erfterm = erf(arg1*def1) + erf(arg2*def1)
              expterm = exp(-0.25d0*r2/asqpt)
              fgb1 = 0.5d0*erfterm/s1
              if (i .eq. k) fgb1 = 2.0d0*fgb1
              fgb2 = screen*rb2*def2*expterm
            else
              fgb1 = 1.0d0/sqrt(r2 + 1.25d0*rb2)
              expterm = exp(-0.25d0*r2/rb2)
              fgb2 = irb* expterm
            end if
            e = fik*(fgb1 + fgb2)
            nes = nes + 1
            if (i .eq. k) then
              epol = epol + 0.5d0*e
              aepol(i) = aepol(i) + 0.5d0*e
            else
              epol = epol + e
              aepol(i) = aepol(i) + 0.5d0*e
              aepol(k) = aepol(k) + 0.5d0*e
            end if
          end if
        end do
      end do
      return
      end
      
