c
c
c     ###################################################
c     ##  COPYRIGHT (C)  1999  by  Jay William Ponder  ##
c     ##              All Rights Reserved              ##
c     ###################################################
c
c     ##############################################################
c     ##                                                          ##
c     ##  subroutine ewald  --  Ewald summation potential energy  ##
c     ##                                                          ##
c     ##############################################################
c
c
c     "ewald" calculates the Ewald summation interaction energy
c
c
      subroutine ewald
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'bound.i'
      include 'cell.i'
      include 'charge.i'
      include 'chgpot.i'
      include 'couple.i'
      include 'cutoff.i'
      include 'energi.i'
      include 'group.i'
      include 'math.i'
      include 'pme.i'
      include 'units.i'
      include 'usage.i'
      integer i,j,k,ii,kk,skip(maxatm)
      real*8 e,f,fi,fik,fs
      real*8 r,r2,rew,fgrp,cut2
      real*8 xi,yi,zi,xr,yr,zr
      real*8 erfc,erfterm,scaleterm
      logical proceed,iuse
c
c
c     zero out the Ewald summation interaction energy
c
      ew = 0.0d0
      if (nion .eq. 0)  return
c
c     zero out the list of atoms to be skipped
c
      do i = 1, n
         skip(i) = 0
      end do
c
c     set conversion factor, cutoff and scaling coefficients
c
      f = electric / dielec
      cut2 = pmecut * pmecut
      scaleterm = 1.0d0 - 1.0d0/chgscale
c
c     compute the Ewald self-energy term over all the atoms
c
      fs = -electric * aewald / sqrtpi
      do ii = 1, nion
         e = fs * pchg(ii)**2
         ew = ew + e
      end do
c
c     compute the reciprocal space part of the Ewald summation
c
      call epme (e)
      ew = ew + e
c
c     compute the direct space portion of the Ewald summation
c
      do ii = 1, nion-1
         i = iion(ii)
         iuse = use(i)
         skip(i) = i
         do j = 1, n12(i)
            skip(i12(j,i)) = i * chg12use
         end do
         do j = 1, n13(i)
            skip(i13(j,i)) = i * chg13use
         end do
         do j = 1, n14(i)
            skip(i14(j,i)) = i * chg14use
         end do
         xi = x(i)
         yi = y(i)
         zi = z(i)
         fi = f * pchg(ii)
         do kk = ii+1, nion
            k = iion(kk)
c
c     decide whether to compute the current interaction
c
            proceed = .true.
            if (use_group)  call groups (proceed,fgrp,2,i,k,0,0,0)
            if (proceed)  proceed = (iuse .or. use(k))
c
c     compute the energy contribution for this interaction
c
            if (proceed) then
               xr = xi - x(k)
               yr = yi - y(k)
               zr = zi - z(k)
               if (use_image)  call image (xr,yr,zr,0)
               r2 = xr*xr + yr*yr + zr*zr
               if (r2 .le. cut2) then
                  r = sqrt(r2)
                  fik = fi * pchg(kk)
                  rew = aewald * r
                  erfterm = erfc (rew)
                  e = fik * erfterm / r
c
c     adjust the direct sum energy for neglected interactions
c
                  if (skip(k) .eq. i) then
                     e = e - fik/r
                  else if (skip(k) .eq. -i) then
                     e = e - (fik/r)*scaleterm
                  end if
c
c     scale the interaction based on its group membership
c
                  if (use_group)  e = e * fgrp
c
c     increment the overall Ewald summation energy component
c
                  ew = ew + e
               end if
            end if
         end do
      end do
c
c     for periodic boundary conditions with large cutoffs
c     neighbors must be found by the replicates method
c
      if (.not. use_replica)  return
c
c     calculate direct space portion involving other unit cells
c
      do ii = 1, nion
         i = iion(ii)
         iuse = use(i)
         xi = x(i)
         yi = y(i)
         zi = z(i)
         fi = f * pchg(ii)
         do kk = ii, nion
            k = iion(kk)
c
c     decide whether to compute the current interaction
c
            proceed = .true.
            if (use_group)  call groups (proceed,fgrp,2,i,k,0,0,0)
            if (proceed)  proceed = (iuse .or. use(k))
c
c     compute the energy contribution for this interaction
c
            if (proceed) then
               do j = 1, ncell
                  xr = xi - x(k)
                  yr = yi - y(k)
                  zr = zi - z(k)
                  call image (xr,yr,zr,j)
                  r2 = xr*xr + yr*yr + zr*zr
                  if (r2 .le. cut2) then
                     r = sqrt(r2)
                     fik = fi * pchg(kk)
                     rew = aewald * r
                     erfterm = erfc (rew)
                     e = fik * erfterm / r
c
c     scale the interaction based on its group membership
c
                     if (use_group)  e = e * fgrp
c
c     increment the overall Ewald summation energy component;
c     interaction of an atom with its own image counts half
c
                     if (i .eq. k)  e = 0.5d0 * e
                     ew = ew + e
                  end if
               end do
            end if
         end do
      end do
      return
      end
c
c
c     #############################################################
c     ##                                                         ##
c     ##  subroutine epme  --  PM Ewald reciprocal space energy  ##
c     ##                                                         ##
c     #############################################################
c
c
c     "epme" computes the reciprocal space energy for a particle
c     mesh Ewald summation over partial charges
c
c     literature reference:
c
c     U. Essmann, L. Perera, M. L Berkowitz, T. Darden, H. Lee and
c     L. G. Pedersen, "A Smooth Particle Mesh Ewald Method", Journal
c     of Chemical Physics, 103, 8577-8593 (1995)
c
c
      subroutine epme (e)
      implicit none
      include 'sizes.i'
      include 'atoms.i'
      include 'boxes.i'
      include 'charge.i'
      include 'math.i'
      include 'pme.i'
      include 'units.i'
      integer i,j,k,m,ii
      integer i0,j0,k0,it1,it2,it3
      integer k1,k2,k3,m1,m2,m3
      integer nf1,nf2,nf3,nff,npoint
      integer ndim1,ndim2,ndim3
      integer ntable,nwork,ngrid
      real*8 xi,yi,zi,w,t3,t2,t1
      real*8 e,msq,struc2,select
      real*8 chg,term,expterm
      real*8 denom,pterm,volterm
      real*8 mhat1,mhat2,mhat3,work(2*maxfft)
      real*8 fr1(maxatm),fr2(maxatm),fr3(maxatm)
      real*8 theta1(maxorder,maxatm),theta2(maxorder,maxatm)
      real*8 theta3(maxorder,maxatm)
      real*8 dtheta1(maxorder,maxatm),dtheta2(maxorder,maxatm)
      real*8 dtheta3(maxorder,maxatm)
c
c
c     initialization of some array dimensions
c
      ndim1 = 2*(nfft1/2) + 1
      ndim2 = 2*(nfft2/2) + 1
      ndim3 = 2*(nfft3/2) + 1
      ntable = maxtable
      nwork = maxfft
      ngrid = maxgrid
      select = sqrt(electric)
c
c     zero out the 3-dimensional charge grid array
c
      do k = 1, ndim3
         do j = 1, ndim2
            do i = 1, ndim1
               qgrid(1,i,j,k) = 0.0d0
               qgrid(2,i,j,k) = 0.0d0
            end do
         end do
      end do
c
c     get the scaled and shifted fractional coordinates
c
      do ii = 1, nion
         i = iion(ii)
         xi = x(i)
         yi = y(i)
         zi = z(i)
         w = xi*recip(1,1) + yi*recip(2,1) + zi*recip(3,1)
         fr1(i) = nfft1 * (w-dble(nint(w))+0.5d0)
         w = xi*recip(1,2) + yi*recip(2,2) + zi*recip(3,2)
         fr2(i) = nfft2 * (w-dble(nint(w))+0.5d0)
         w = xi*recip(1,3) + yi*recip(2,3) + zi*recip(3,3)
         fr3(i) = nfft3 * (w-dble(nint(w))+0.5d0)
c
c     get the B-spline coefficients and their first derivatives
c
         w = fr1(i) - dble(int(fr1(i)))
         call bspline (w,order,theta1(1,i),dtheta1(1,i))
         w = fr2(i) - dble(int(fr2(i)))
         call bspline (w,order,theta2(1,i),dtheta2(1,i))
         w = fr3(i) - dble(int(fr3(i)))
         call bspline (w,order,theta3(1,i),dtheta3(1,i))
      end do
c
c     fill the 3-dimensional charge grid array
c
      do ii = 1, nion
         m = iion(ii)
         chg = select * pchg(ii)
         k0 = int(fr3(m)) - order
         do it3 = 1, order
            t3 = theta3(it3,m) * chg
            k0 = k0 + 1
            k = k0 + 1 + (nfft3-sign(nfft3,k0))/2
            j0 = int(fr2(m)) - order
            do it2 = 1, order
               t2 = theta2(it2,m)
               term = t2 * t3
               j0 = j0 + 1
               j = j0 + 1 + (nfft2-sign(nfft2,j0))/2
               i0 = int(fr1(m)) - order
               do it1 = 1, order
                  t1 = theta1(it1,m)
                  i0 = i0 + 1
                  i = i0 + 1 + (nfft1-sign(nfft1,i0))/2
                  qgrid(1,i,j,k) = qgrid(1,i,j,k) + term*t1
               end do
            end do
         end do
      end do
c
c     perform the 3-D FFT forward transformation
c
      call fftfront (nfft1,nfft2,nfft3,ngrid,ngrid,ngrid,
     &               ntable,nwork,table,work,qgrid)
c
c     sum the reciprocal space interaction energy components
c
      npoint = nfft1 * nfft2 * nfft3
      pterm = (pi/aewald)**2
      volterm = pi * volbox
      nff = nfft1 * nfft2
      nf1 = (nfft1+1) / 2
      nf2 = (nfft2+1) / 2
      nf3 = (nfft3+1) / 2
      e = 0.0d0
      do i = 1, npoint-1
         k3 = i/nff + 1
         j = i - (k3-1)*nff
         k2 = j/nfft1 + 1
         k1 = j - (k2-1)*nfft1 + 1
         m1 = k1 - 1
         if (k1 .gt. nf1)  m1 = m1 - nfft1
         m2 = k2 - 1
         if (k2 .gt. nf2)  m2 = m2 - nfft2
         m3 = k3 - 1
         if (k3 .gt. nf3)  m3 = m3 - nfft3
         mhat1 = recip(1,1)*m1 + recip(1,2)*m2 + recip(1,3)*m3
         mhat2 = recip(2,1)*m1 + recip(2,2)*m2 + recip(2,3)*m3
         mhat3 = recip(3,1)*m1 + recip(3,2)*m2 + recip(3,3)*m3
         msq = mhat1**2 + mhat2**2 + mhat3**2
         term = -pterm * msq
         expterm = 0.0d0
         if (term .gt. -50.0d0) then
            denom = volterm*msq*bsmod1(k1)*bsmod2(k2)*bsmod3(k3)
            expterm = exp(term) / denom
            struc2 = qgrid(1,k1,k2,k3)**2 + qgrid(2,k1,k2,k3)**2
            e = e + expterm*struc2
         end if
         qgrid(1,k1,k2,k3) = expterm * qgrid(1,k1,k2,k3)
         qgrid(2,k1,k2,k3) = expterm * qgrid(2,k1,k2,k3)
      end do
      e = 0.5d0 * e
      return
      end
