c
c
c     ###################################################
c     ##  COPYRIGHT (C)  1990  by  Jay William Ponder  ##
c     ##              All Rights Reserved              ##
c     ###################################################
c
c     ##################################################################
c     ##                                                              ##
c     ##  subroutine lattice  --  setup periodic boundary conditions  ##
c     ##                                                              ##
c     ##################################################################
c
c
c     "lattice" stores the periodic box dimensions, finds angles
c     to be used in computing fractional coordinates, and decides
c     between images and replicates for the boundary conditions
c
c
      subroutine lattice
      implicit none
      include 'sizes.i'
      include 'bound.i'
      include 'boxes.i'
      include 'cell.i'
      include 'cutoff.i'
      include 'inform.i'
      include 'iounit.i'
      include 'math.i'
      include 'potent.i'
      integer i,j,k
      integer nxcell,nycell,nzcell
      real*8 ar1,ar2,ar3
      real*8 br1,br2,br3
      real*8 cr1,cr2,cr3
      real*8 xlimit,ylimit,zlimit
      real*8 maximage,maxcut,alpha_cos
c
c
c     only analyze lattice if periodic boundaries are in use
c
      if (.not. use_bounds)  return
c
c     set the half width values for the periodic box
c
      xbox2 = 0.5d0 * xbox
      ybox2 = 0.5d0 * ybox
      zbox2 = 0.5d0 * zbox
      if (octahedron)  box34 = 0.75d0 * xbox
c
c     set values used in computation of fractional coordinates
c
      if (orthogonal .or. octahedron) then
         alpha_cos = 0.0d0
         beta_sin = 1.0d0
         beta_cos = 0.0d0
         gamma_sin = 1.0d0
         gamma_cos = 0.0d0
         beta_term = 0.0d0
         gamma_term = 1.0d0
      else if (monoclinic) then
         alpha_cos = 0.0d0
         beta_sin = sin(beta/radian)
         beta_cos = cos(beta/radian)
         gamma_sin = 1.0d0
         gamma_cos = 0.0d0
         beta_term = 0.0d0
         gamma_term = beta_sin
      else if (triclinic) then
         alpha_cos = cos(alpha/radian)
         beta_sin = sin(beta/radian)
         beta_cos = cos(beta/radian)
         gamma_sin = sin(gamma/radian)
         gamma_cos = cos(gamma/radian)
         beta_term = (alpha_cos - beta_cos*gamma_cos) / gamma_sin
         gamma_term = sqrt(1.0d0 - beta_cos**2 - beta_term**2)
      end if
c
c     determine the volume of the periodic box
c
      if (orthogonal) then
         volbox = xbox * ybox * zbox
      else if (monoclinic) then
         volbox = beta_sin * xbox * ybox * zbox
      else if (triclinic) then
         volbox = (gamma_sin*gamma_term) * xbox * ybox * zbox
      else if (octahedron) then
         volbox = 0.5d0 * xbox * ybox * zbox
      end if
c
c     place the reciprocal lattice vectors into columns
c
      ar1 = xbox
      ar2 = 0.0d0
      ar3 = 0.0d0
      br1 = ybox * gamma_cos
      br2 = ybox * gamma_sin
      br3 = 0.0d0
      cr1 = zbox * beta_cos
      cr2 = zbox * beta_term
      cr3 = zbox * gamma_term
      recip(1,1) = (br2*cr3 - cr2*br3) / volbox
      recip(2,1) = (br3*cr1 - cr3*br1) / volbox
      recip(3,1) = (br1*cr2 - cr1*br2) / volbox
      recip(1,2) = (cr2*ar3 - ar2*cr3) / volbox
      recip(2,2) = (cr3*ar1 - ar3*cr1) / volbox
      recip(3,2) = (cr1*ar2 - ar1*cr2) / volbox
      recip(1,3) = (ar2*br3 - br2*ar3) / volbox
      recip(2,3) = (ar3*br1 - br3*ar1) / volbox
      recip(3,3) = (ar1*br2 - br1*ar2) / volbox
c
c     find the maximum sphere radius inscribed in periodic box
c
      if (orthogonal) then
         xlimit = xbox2
         ylimit = ybox2
         zlimit = zbox2
      else if (monoclinic) then
         xlimit = xbox2 * beta_sin
         ylimit = ybox2
         zlimit = zbox2 * beta_sin
      else if (triclinic) then
         xlimit = xbox2 * beta_sin * gamma_sin
         ylimit = ybox2 * gamma_sin
         zlimit = zbox2 * beta_sin
      else if (octahedron) then
         xlimit = (sqrt(3.0d0)/4.0d0) * xbox
         ylimit = xlimit
         zlimit = xlimit
      end if
      maximage = min(xlimit,ylimit,zlimit)
c
c     set default values for the energy term cutoffs
c
      if (chgcut.eq.0.0d0 .or. chgcut.ge.1000000.0d0)  chgcut = 9.0d0
      if (dplcut.eq.0.0d0 .or. dplcut.ge.1000000.0d0)  dplcut = 9.0d0
      if (vdwcut.eq.0.0d0 .or. vdwcut.ge.1000000.0d0)  vdwcut = 9.0d0
c
c     use replicate method to handle cutoffs too large for images
c
      maxcut = 0.0d0
      if (use_vdw)  maxcut = max(maxcut,vdwcut)
      if (use_charge)  maxcut = max(maxcut,chgcut)
      if (use_dipole)  maxcut = max(maxcut,dplcut)
      if (use_ewald)  maxcut = max(maxcut,pmecut)
      if (maxcut .le. maximage) then
         use_replica = .false.
      else
         use_replica = .true.
      end if
c
c     truncated octahedron cannot use the replicates method
c
      if (octahedron .and. use_replica) then
         write (iout,10)
   10    format (/,' LATTICE  --  Truncated Octahedron',
     &              ' cannot be Replicated')
         call fatal
      end if
c
c     find the number of replicates for orthogonal lattice
c
      nxcell = int(maxcut/xlimit)
      nycell = int(maxcut/ylimit)
      nzcell = int(maxcut/zlimit)
      if (maxcut .gt. nxcell*xlimit)  nxcell = nxcell + 1
      if (maxcut .gt. nycell*ylimit)  nycell = nycell + 1
      if (maxcut .gt. nzcell*zlimit)  nzcell = nzcell + 1
      xcell = nxcell * xbox
      ycell = nycell * ybox
      zcell = nzcell * zbox
c
c     check the total number of replicated unit cells
c
      ncell = nxcell*nycell*nzcell - 1
      if (ncell .gt. maxcell) then
         write (iout,20)
   20    format (/,' LATTICE  --  Increase MAXCELL or Decrease',
     &              ' the Interaction Cutoffs')
         call fatal
      end if
c
c     assign indices to the required cell replicates
c
      ncell = 0
      do k = 0, nzcell-1
         do j = 0, nycell-1
            do i = 0, nxcell-1
               if (k.ne.0 .or. j.ne.0 .or. i.ne.0) then
                  ncell = ncell + 1
                  icell(1,ncell) = i
                  icell(2,ncell) = j
                  icell(3,ncell) = k
               end if
            end do
         end do
      end do
c
c     set the half width values for the full replicated cell
c
      xcell2 = 0.5d0 * xcell
      ycell2 = 0.5d0 * ycell
      zcell2 = 0.5d0 * zcell
c
c     print a message indicating the number of replicates used
c
      if (debug) then
         write (iout,30)  nxcell,nycell,nzcell
   30    format (/,' LATTICE  --  Periodicity using ',i2,' x',i2,
     &              ' x',i2,' Set of Cell Replicates')
      end if
      return
      end
