c
c
c     ###################################################
c     ##  COPYRIGHT (C)  1999  by  Jay William Ponder  ##
c     ##              All Rights Reserved              ##
c     ###################################################
c
c     ##############################################################
c     ##                                                          ##
c     ##  subroutine lbfgs  --  limited memory BFGS optimization  ##
c     ##                                                          ##
c     ##############################################################
c
c
c     "lbfgs" is a limited memory BFGS quasi-newton nonlinear
c     optimization routine
c
c     literature references:
c
c     J. Nocedal, "Updating Quasi-Newton Matrices with Limited
c     Storage", Mathematics of Computation, 35, 773-782 (1980)
c
c     D. C. Lui and J. Nocedal, "On the Limited Memory BFGS Method
c     for Large Scale Optimization", Mathematical Programming,
c     45, 503-528 (1989)
c
c
      subroutine lbfgs (nvar,x,minimum,grdmin,fgvalue,writeout)
      implicit none
      include 'sizes.i'
      include 'iounit.i'
      include 'keys.i'
      include 'linmin.i'
      include 'math.i'
      include 'minima.i'
      include 'potent.i'
      include 'output.i'
      include 'scales.i'
      include 'solute.i'
      integer maxsav
      parameter (maxsav=15)
      integer i,j,k,m,npt,pt,cp
      integer nvar,muse,next,info
      integer niter,ncalls,nint
      real*8 fgvalue,stp
      real*8 f,fold,fmove,xmove
      real*8 gnorm,grms,rms
      real*8 sg,snorm,cosang,angle
      real*8 minimum,grdmin
      real*8 xtol,ftol,beta
      real*8 ys,sq,yr,yy
      real*8 x(maxvar),g(maxvar),diag(maxvar)
      real*8 s(maxvar*maxsav),y(maxvar*maxsav)
      real*8 w(maxvar+2*maxsav)
      character*9 status
      character*20 keyword
      character*80 record
      logical done,repeat
      external fgvalue,writeout
c
c
c     initialize some values to be used below
c
      if (nvar .gt. maxvar) then
         write (iout,10)
   10    format (/,' LBFGS  --  Too many Parameters,',
     &              ' Increase value of MAXVAR')
         return
      end if
      ncalls = 0
      rms = sqrt(dble(nvar))
      if (coordtype .eq. 'cartesian') then
         rms = rms / sqrt(3.0d0)
      else if (coordtype .eq. 'rigidbody') then
         rms = rms / sqrt(6.0d0)
      end if
c
c     set default parameters for the optimization
c
      m = min(nvar,maxsav)
      if (fctmin .eq. 0.0d0)  fctmin = -1000000.0d0
      if (maxiter .eq. 0)  maxiter = 1000000
      if (iprint .lt. 0)  iprint = 1
      if (iwrite .lt. 0)  iwrite = 1
c
c     set default parameters for the line search
c
      if (cappa .eq. 0.0d0)  cappa = 0.9d0
      if (stpmin .eq. 0.0d0)  stpmin = 1.0d-16
      if (stpmax .eq. 0.0d0)  stpmax = 1.0d16
      intmax = 20
      xtol = 1.0d-16
      ftol = 1.0d-4
c
c     search the keywords for optimization parameters
c
      do i = 1, nkey
         next = 1
         record = keyline(i)
         call gettext (record,keyword,next)
         call upcase (keyword)
         if (keyword(1:14) .eq. 'LBFGS-VECTORS ') then
            read (record(next:80),*,err=20)  m
         else if (keyword(1:7) .eq. 'FCTMIN ') then
            read (record(next:80),*,err=20)  fctmin
         else if (keyword(1:8) .eq. 'MAXITER ') then
            read (record(next:80),*,err=20)  maxiter
         else if (keyword(1:9) .eq. 'PRINTOUT ') then
            read (record(next:80),*,err=20)  iprint
         else if (keyword(1:9) .eq. 'WRITEOUT ') then
            read (record(next:80),*,err=20)  iwrite
         else if (keyword(1:8) .eq. 'STEPMAX ') then
            read (record(next:80),*,err=20)  stpmax
         else if (keyword(1:8) .eq. 'STEPMIN ') then
            read (record(next:80),*,err=20)  stpmin
         else if (keyword(1:6) .eq. 'CAPPA ') then
            read (record(next:80),*,err=20)  cappa
         else if (keyword(1:7) .eq. 'INTMAX ') then
            read (record(next:80),*,err=20)  intmax
         end if
   20    continue
      end do
c
c     check the number of saved correction vectors
c
      if (m.le.0 .or. m.gt.min(nvar,maxsav)) then
         m = min(nvar,maxsav)
         write (iout,30)  m
   30    format (/,' LBFGS  --  Number of Saved Vectors Set to'
     &              ' the Maximum of',i5)
      end if
c
c     evaluate the function and get the initial gradient
c
      done = .false.
      niter = 0
      pt = 0
      ncalls = 1
      f = fgvalue (x,g)
      fold = f
      grms = 0.0d0
      do i = 1, nvar
         grms = grms + (g(i)*scale(i))**2
      end do
      grms = sqrt(grms) / rms
c
c     print header information prior to first iteration
c
      if (iprint .gt. 0) then
         write (iout,40)
   40    format (/,' Limited Memory BFGS Quasi-Newton Optimization :')
         write (iout,50)
   50    format (/,' QN Iter    F Value      G RMS     F Move'
     &              '    X Move    Angle  FG Call  Comment',/)
         if (f.lt.1.0d7 .and. f.gt.-1.0d6 .and. grms.lt.1.0d5) then
            write (iout,60)  niter,f,grms,ncalls
   60       format (i6,f13.4,f11.4,30x,i7)
         else
            write (iout,70)  niter,f,grms,ncalls
   70       format (i6,d13.4,d11.4,30x,i7)
         end if
      end if
c
c     beginning of a new L-BFGS iteration cycle
c
      dowhile (.not. done)
         niter = niter + 1
         muse = min(niter-1,m)
         stp = 1.0d0
c
c     if using GB/SA solvation, update Born radii occasionally
c
         if (use_gbsa) then
            if (mod(niter,50) .eq. 0) then
               reborn = 1
               call born
               f = fgvalue (x,g)
            end if
            reborn = 0
         end if
c
c     estimate an appropriate stepsize for the first iteration
c
         if (niter .eq. 1) then
            gnorm = 0.0d0
            do i = 1, nvar
               diag(i) = 1.0d0
               s(i) = -g(i) * diag(i)
               gnorm = gnorm + g(i)*g(i)
            end do
            gnorm = sqrt(gnorm)
            stp = 1.0d0 / gnorm
c
c     compute -Hg, diagonal scaling vector and search direction
c
         else
            do i = 1, nvar
               diag(i) = ys / yy
               if (diag(i) .le. 0.0d0) then
                  write (iout,80)
   80             format (/,' LBFGS  --  Diagonal Elements of Inverse',
     &                       ' Hessian must be Positive')
               end if
            end do
            cp = pt
            if (pt .eq. 0)  cp = m
            w(nvar+cp) = 1.0d0 / ys
            do i = 1, nvar
               w(i) = -g(i)
            end do
            cp = pt
            do k = 1, muse
               cp = cp - 1
               if (cp .eq. -1)  cp = m - 1
               sq = 0.0d0
               do i = 1, nvar
                  sq = sq + s(cp*nvar+i)*w(i)
               end do
               j = nvar + cp + 1
               w(j+m) = w(j) * sq
               do i = 1, nvar
                  w(i) = w(i) - w(j+m)*y(cp*nvar+i)
               end do
            end do
            do i = 1, nvar
               w(i) = diag(i) * w(i)
            end do
            do k = 1, muse
               yr = 0.0d0
               do i = 1, nvar
                  yr = yr + y(cp*nvar+i)*w(i)
               end do
               j = nvar + cp + 1
               beta = w(j) * yr
               do i = 1, nvar
                  w(i) = w(i) + s(cp*nvar+i)*(w(j+m)-beta)
               end do
               cp = cp + 1
               if (cp .eq. m)  cp = 0
            end do
            do i = 1, nvar
               s(pt*nvar+i) = w(i)
            end do
         end if
c
c     store the gradient and get its angle with search direction
c
         sg = 0.0d0
         snorm = 0.0d0
         gnorm = 0.0d0
         do i = 1, nvar
            w(i) = g(i)
            sg = sg + s(i)*g(i)
            snorm = snorm + s(i)*s(i)
            gnorm = gnorm + g(i)*g(i)
         end do
         snorm = sqrt(snorm)
         gnorm = sqrt(gnorm)
         cosang = -sg / (snorm * gnorm)
         cosang = min(1.0d0,max(-1.0d0,cosang))
         angle = radian * acos(cosang)
c
c     minimize along the search direction via a line search
c
         info = 0
         nint = 0
         repeat = .true.
         dowhile (repeat)
            call lbfgs1 (nvar,x,f,g,s(pt*nvar+1),stp,
     &                    xtol,ftol,info,nint,diag)
            if (info .eq. -1) then
               ncalls = ncalls + 1
               f = fgvalue (x,g)
            else
               repeat = .false.
            end if
         end do
c
c     get the sizes of the moves made during this iteration
c
         fmove = fold - f
         fold = f
         xmove = 0.0d0
         do i = 1, nvar
            xmove = xmove + (s(i)/scale(i))**2
         end do
         xmove = stp * sqrt(xmove) / rms
         if (coordtype .eq. 'internal') then
            xmove = radian * xmove
         end if
c
c     update variables based on results of this iteration
c
         ys = 0.0d0
         yy = 0.0d0
         npt = pt * nvar
         do i = 1, nvar
            k = npt + i
            s(k) = stp * s(k)
            y(k) = g(i) - w(i)
            ys = ys + y(k)*s(k)
            yy = yy + y(k)*y(k)
         end do
         pt = pt + 1
         if (pt .eq. m)  pt = 0
c
c     compute the rms gradient per optimization parameter
c
         grms = 0.0d0
         do i = 1, nvar
            grms = grms + (g(i)*scale(i))**2
         end do
         grms = sqrt(grms) / rms
c
c     check for errors during the line search
c
         if (info .eq. 1) then
            status = ' Success '
         else if (info .eq. 0) then
            status = 'InputErr '
            done = .true.
         else if (info .eq. 2) then
            status = 'TrustFail'
            done = .true.
         else if (info .eq. 3) then
            status = 'MaxLinItr'
            done = .true.
         else if (info .eq. 4) then
            status = ' StepMin '
            done = .true.
         else if (info .eq. 5) then
            status = ' StepMax '
            done = .true.
         else if (info .eq. 6) then
            status = 'RoundOff '
            done = .true.
         end if
c
c     test for too many total iterations
c
         if (niter .ge. maxiter) then
            status = 'IterLimit'
            done = .true.
         end if
c
c     test the normal termination criteria
c
         if (f .le. fctmin) then
            status = 'SmallFct '
            done = .true.
         end if
         if (grms .le. grdmin) then
            status = 'SmallGrad'
            done = .true.
         end if
c
c     print intermediate results every few iterations
c
         if (iprint .gt. 0) then
            if (done .or. mod(niter,iprint).eq.0) then
               if (f.lt.1.0d7 .and. f.gt.-1.0d6 .and.
     &             grms.lt.1.0d5 .and. fmove.lt.1.0d5) then
                  write (iout,90)  niter,f,grms,fmove,xmove,
     &                              angle,ncalls,status
   90             format (i6,f13.4,f11.4,f11.4,f10.4,f9.2,i7,3x,a9)
               else
                  write (iout,100)  niter,f,grms,fmove,xmove,
     &                              angle,ncalls,status
  100             format (i6,d13.4,d11.4,d11.4,f10.4,f9.2,i7,3x,a9)
               end if
            end if
         end if
c
c     write intermediate results every few iterations
c
         if (iwrite .gt. 0) then
            if (done .or. mod(niter,iwrite).eq.0) then
               call writeout (x,niter)
            end if
         end if
      end do
c
c     set final value of the objective function
c
      minimum = f
      if (iprint .gt. 0) then
         if (status.eq.'SmallGrad' .or. status.eq.'SmallFct ') then
            write (iout,110)  status
  110       format (/,' LBFGS  --  Normal Termination due to ',a9)
         else
            write (iout,120)  status
  120       format (/,' LBFGS  --  Incomplete Convergence due to ',a9)
         end if
      end if
      return
      end
c
c
c     #################################################################
c     ##                                                             ##
c     ##  subroutine lbfgs1  --  line search for lbfgs optimization  ##
c     ##                                                             ##
c     #################################################################
c
c
      subroutine lbfgs1 (nvar,x,f,g,s,stp,xtol,ftol,info,nint,diag)
      include 'sizes.i'
      include 'linmin.i'
      integer i,nvar,info,nint,infoc
      real*8 f,stp,xtol,ftol
      real*8 dg,dgm,dginit,dgtest
      real*8 dgx,dgxm,dgy,dgym
      real*8 finit,ftest1,stx,sty
      real*8 fm,fx,fxm,fy,fym
      real*8 stmin,stmax,xtrapf
      real*8 width,width1
      real*8 x(maxvar),g(maxvar)
      real*8 s(maxvar),diag(maxvar)
      logical brackt,stage1
c
c
c     check the validity of the input parameters
c
      if (info .ne. -1) then
         infoc = 1
         if (nvar.le.0 .or. stp.le.0.0d0 .or. intmax.le.0) return
         if (min(xtol,ftol,cappa) .lt. 0.0d0)  return
         if (stpmin.lt.0.0d0 .or. stpmax.lt.stpmin)  return
c
c     compute the initial gradient in the search direction
c
         dginit = 0.0d0
         do i = 1, nvar
            dginit = dginit + g(i)*s(i)
         end do
         if (dginit .ge. 0.0d0)  return
c
c     initialize the local variables
c
         xtrapf = 4.0d0
         brackt = .false.
         stage1 = .true.
         nint = 0
         finit = f
         dgtest = ftol * dginit
         width = stpmax - stpmin
         width1 = width / 0.5d0
         do i = 1, nvar
            diag(i) = x(i)
         end do
c
c     the variables stx, fx, dgx contain the values of the step,
c     function, and directional derivative at the best step.
c     the variables sty, fy, dgy contain the value of the step,
c     function, and derivative at the other endpoint of
c     the interval of uncertainty.
c     the variables stp, f, dg contain the values of the step,
c     function, and derivative at the current step.
c
         stx = 0.0d0
         fx = finit
         dgx = dginit
         sty = 0.0d0
         fy = finit
         dgy = dginit
c
c     compute the directional derivative
c
      else
         info = 0
         nint = nint + 1
         dg = 0.0d0
         do i = 1, nvar
            dg = dg + g(i)*s(i)
         end do
         ftest1 = finit + stp*dgtest
c
c     test for error conditions and convergence
c
         if ((brackt .and. (stp.le.stmin .or. stp.ge.stmax))
     &         .or. infoc.eq.0)  info = 6
         if (stp.eq.stpmax .and. f.le.ftest1 .and. dg.le.dgtest)
     &         info = 5
         if (stp.eq.stpmin .and. (f.gt.ftest1 .or. dg.ge.dgtest))
     &         info = 4
         if (nint .ge. intmax)  info = 3
         if (brackt .and. stmax-stmin.le.xtol*stmax)  info = 2
         if (f.le.ftest1 .and. abs(dg).le.cappa*(-dginit))  info = 1
c
c     check for normal or error termination
c
         if (info .ne. 0)  return
c
c     in the first stage we seek a step for which the modified
c     function has a nonpositive value and nonnegative derivative
c
         if (stage1 .and. f.le.ftest1 .and.
     &          dg.ge.min(ftol,cappa)*dginit)  stage1 = .false.
c
c     a modified function is used to predict the step only if there
c     is no step for which the modified function has a nonpositive
c     function value and nonnegative derivative, and if a lower
c     function value has been obtained but with insufficient decrease
c
         if (stage1 .and. f.le.fx .and. f.gt.ftest1) then
c
c     define the modified function and derivative values
c
            fm = f - stp*dgtest
            fxm = fx - stx*dgtest
            fym = fy - sty*dgtest
            dgm = dg - dgtest
            dgxm = dgx - dgtest
            dgym = dgy - dgtest
c
c     update the interval of uncertainty and compute the new step
c
            call lbfgs2 (stx,fxm,dgxm,sty,fym,dgym,stp,fm,dgm,
     &                        brackt,stmin,stmax,infoc)
c
c     reset the function and gradient values
c
            fx = fxm + stx*dgtest
            fy = fym + sty*dgtest
            dgx = dgxm + dgtest
            dgy = dgym + dgtest
         else
c
c     update the interval of uncertainty and compute the new step
c
            call lbfgs2 (stx,fx,dgx,sty,fy,dgy,stp,f,dg,
     &                      brackt,stmin,stmax,infoc)
         end if
c
c     force a sufficient decrease in the interval of uncertainty
c
         if (brackt) then
            if (abs(sty-stx) .ge. 0.66d0*width1)
     &         stp = stx + 0.5d0*(sty-stx)
            width1 = width
            width = abs(sty-stx)
         end if
      end if
c
c     set the minimum and maximum steps to correspond
c     to the present interval of uncertainty
c
      if (brackt) then
         stmin = min(stx,sty)
         stmax = max(stx,sty)
      else
         stmin = stx
         stmax = stp + xtrapf*(stp-stx)
      end if
c
c     force the step to be within bounds on the step size
c
      stp = max(stp,stpmin)
      stp = min(stp,stpmax)
c
c     if an unusual termination is to occur then let
c     stp be the lowest point obtained so far
c
      if ((brackt .and. (stp.le.stmin .or. stp.ge.stmax)) .or.
     &    nint.ge.intmax-1 .or. infoc.eq.0  .or.
     &    (brackt .and. stmax-stmin.le.xtol*stmax))  stp = stx
c
c     evaluate the function and gradient after a step
c
      info = -1
      do i = 1, nvar
         x(i) = diag(i) + stp*s(i)
      end do
      return
      end
c
c
c     #################################################################
c     ##                                                             ##
c     ##  subroutine lbfgs2  --  trust region for lbfgs line search  ##
c     ##                                                             ##
c     #################################################################
c
c
      subroutine lbfgs2 (stx,fx,dx,sty,fy,dy,stp,fp,dp,
     &                    brackt,stpmin,stpmax,info)
      integer info
      real*8 stx,fx,dx,sty,fy,dy,stp,fp,dp
      real*8 stpmin,stpmax,gamma,theta
      real*8 p,q,r,s,sgnd,stpc,stpf,stpq
      logical brackt,bound
c
c
c     check the input parameters for errors
c
      info = 0
      if ((brackt .and. (stp.le.min(stx,sty)
     &               .or.stp.ge.max(stx,sty))) .or.
     &    dx*(stp-stx).ge.0.0d0 .or. stpmax.lt.stpmin)  return
c
c     determine if the derivatives have opposite sign
c
      sgnd = dp * (dx/abs(dx))
c
c     first case; a higher function value; the minimum is bracketed;
c     if the cubic step is closer to stx than the quadratic step,
c     the cubic step is taken, else the average of the cubic and
c     quadratic steps is taken
c
      if (fp .gt. fx) then
         info = 1
         bound = .true.
         theta = 3.0d0*(fx-fp)/(stp-stx) + dx + dp
         s = max(abs(theta),abs(dx),abs(dp))
         gamma = s * sqrt((theta/s)**2-(dx/s)*(dp/s))
         if (stp .lt. stx)  gamma = -gamma
         p = (gamma-dx) + theta
         q = ((gamma-dx) + gamma) + dp
         r = p / q
         stpc = stx + r*(stp-stx)
         stpq = stx + ((dx/((fx-fp)/(stp-stx)+dx))/2.0d0) * (stp-stx)
         if (abs(stpc-stx) .lt. abs(stpq-stx)) then
            stpf = stpc
         else
            stpf = stpc + (stpq-stpc)/2.0d0
         end if
         brackt = .true.
c
c     second case; a lower function value and derivatives of
c     opposite sign; the minimum is bracketed; if the cubic
c     step is closer to stx than the quadratic (secant) step,
c     the cubic step is taken, else the quadratic step is taken
c
      else if (sgnd .lt. 0.0) then
         info = 2
         bound = .false.
         theta = 3.0d0*(fx-fp)/(stp-stx) + dx + dp
         s = max(abs(theta),abs(dx),abs(dp))
         gamma = s * sqrt((theta/s)**2-(dx/s)*(dp/s))
         if (stp .gt. stx)  gamma = -gamma
         p = (gamma-dp) + theta
         q = ((gamma-dp) + gamma) + dx
         r = p / q
         stpc = stp + r*(stx-stp)
         stpq = stp + (dp/(dp-dx))*(stx-stp)
         if (abs(stpc-stp) .gt. abs(stpq-stp)) then
            stpf = stpc
         else
            stpf = stpq
         end if
         brackt = .true.
c
c     third case; a lower function value, derivatives of the
c     same sign, and the magnitude of the derivative decreases;
c     the cubic step is only used if the cubic tends to infinity
c     in the direction of the step or if the minimum of the cubic
c     is beyond stp, otherwise the cubic step is defined to be
c     either stpmin or stpmax; the quadratic (secant) step is also
c     computed and if the minimum is bracketed then the the step
c     closest to stx is taken, else the step farthest away is taken
c
      else if (abs(dp) .lt. abs(dx)) then
         info = 3
         bound = .true.
         theta = 3.0d0*(fx-fp)/(stp-stx) + dx + dp
         s = max(abs(theta),abs(dx),abs(dp))
c
c     the case gamma = 0 only arises if the cubic does not tend
c     to infinity in the direction of the step
c
         gamma = s * sqrt(max(0.0d0,(theta/s)**2-(dx/s)*(dp/s)))
         if (stp .gt. stx)  gamma = -gamma
         p = (gamma-dp) + theta
         q = (gamma + (dx-dp)) + gamma
         r = p / q
         if (r.lt.0.0d0 .and. gamma.ne.0.0d0) then
            stpc = stp + r*(stx - stp)
         else if (stp .gt. stx) then
            stpc = stpmax
         else
            stpc = stpmin
         end if
         stpq = stp + (dp/(dp-dx))*(stx-stp)
         if (brackt) then
            if (abs(stp-stpc) .lt. abs(stp-stpq)) then
               stpf = stpc
            else
               stpf = stpq
            end if
         else
            if (abs(stp-stpc) .gt. abs(stp-stpq)) then
               stpf = stpc
            else
               stpf = stpq
            end if
         end if
c
c     fourth case; a lower function value, derivatives of the
c     same sign, and the magnitude of the derivative does
c     not decrease; if the minimum is not bracketed, the step
c     is either stpmin or stpmax, else the cubic step is taken
c
      else
         info = 4
         bound = .false.
         if (brackt) then
            theta = 3.0d0*(fp-fy)/(sty-stp) + dy + dp
            s = max(abs(theta),abs(dy),abs(dp))
            gamma = s * sqrt((theta/s)**2-(dy/s)*(dp/s))
            if (stp .gt. sty)  gamma = -gamma
            p = (gamma-dp) + theta
            q = ((gamma-dp) + gamma) + dy
            r = p / q
            stpc = stp + r*(sty-stp)
            stpf = stpc
         else if (stp .gt. stx) then
            stpf = stpmax
         else
            stpf = stpmin
         end if
      end if
c
c     update the interval of uncertainty
c
      if (fp .gt. fx) then
         sty = stp
         fy = fp
         dy = dp
      else
         if (sgnd .lt. 0.0d0) then
            sty = stx
            fy = fx
            dy = dx
         end if
         stx = stp
         fx = fp
         dx = dp
      end if
c
c     compute the new safegaurded step size
c
      stpf = min(stpmax,stpf)
      stpf = max(stpmin,stpf)
      stp = stpf
      if (brackt .and. bound) then
         if (sty .gt. stx) then
            stp = min(stx+0.66d0*(sty-stx),stp)
         else
            stp = max(stx+0.66d0*(sty-stx),stp)
         end if
      end if
      return
      end
